
"""Meta
"""

import os
from crocodile.core import randstr, str2timedelta, Save, install_n_import, List, Struct
from crocodile.file_management import P, datetime, OPLike, PLike
import time
import logging
import subprocess
import sys
from typing import Union, Any, Optional, Callable, TextIO, BinaryIO, IO, TypeAlias, Literal
from dataclasses import dataclass


_ = IO, TextIO
SHELLS: TypeAlias = Literal["default", "cmd", "powershell", "pwsh", "bash"]  # pwsh.exe is PowerShell (community) and powershell.exe is Windows Powershell (msft)
CONSOLE: TypeAlias = Literal["wt", "cmd"]
MACHINE: TypeAlias = Literal["Windows", "Linux", "Darwin"]


@dataclass
class Scout:
    source_full: P
    source_rel2home: P
    exists: bool
    is_dir: bool
    files: Optional[List[P]]


class Null:
    def __init__(self, return_: Any = 'self'): self.return_ = return_
    def __getattr__(self, item: str) -> 'Null': _ = item; return self if self.return_ == 'self' else self.return_
    def __getitem__(self, item: str) -> 'Null': _ = item; return self if self.return_ == 'self' else self.return_
    def __call__(self, *args: Any, **kwargs: Any) -> 'Null': _ = args, kwargs; return self if self.return_ == 'self' else self.return_
    def __len__(self): return 0
    def __bool__(self): return False
    def __contains__(self, item: str): _ = self, item; return False
    def __iter__(self): return iter([self])


class Log(logging.Logger):  #
    def __init__(self, dialect: Literal["colorlog", "logging", "coloredlogs"] = "colorlog", name: Optional[str] = None, file: bool = False, file_path: OPLike = None, stream: bool = True, fmt: Optional[str] = None, sep: str = " | ",
                 s_level: int = logging.DEBUG, f_level: int = logging.DEBUG, l_level: int = logging.DEBUG, verbose: bool = False,
                 log_colors: Optional[dict[str, str]] = None):
        if name is None:
            name = randstr(noun=True)
            print(f"Logger name not passed. It is recommended to pass a name indicating the owner.")
        super().__init__(name, level=l_level)  # logs everything, finer level of control is given to its handlers
        print(f"Logger `{name}` from `{dialect}` is instantiated with level {l_level}."); self.file_path = file_path  # proper update to this value by self.add_filehandler()
        if dialect == "colorlog":
            import colorlog
            module: Any = colorlog
            processed_fmt: Any = colorlog.ColoredFormatter(fmt or (rf"%(log_color)s" + Log.get_format(sep)), datefmt="%d %H:%M:%S", log_colors=log_colors or {'DEBUG': 'bold_cyan', 'INFO': 'green', 'WARNING': 'yellow', 'ERROR': 'thin_red', 'CRITICAL': 'fg_bold_red,bg_black', })  # see here for format: https://pypi.org/project/colorlog/
        else:
            module = logging
            processed_fmt = logging.Formatter(fmt or Log.get_format(sep))
        if file or file_path: self.add_filehandler(file_path=file_path, fmt=processed_fmt, f_level=f_level)
        if stream: self.add_streamhandler(s_level, fmt=processed_fmt, module=module)
        self.specs = dict(dialect=dialect, name=self.name, file=file, file_path=self.file_path, stream=bool(self.get_shandler()), fmt=fmt, sep=sep, s_level=s_level, f_level=f_level, l_level=l_level, verbose=verbose, log_colors=log_colors)  # # this way of creating relative path makes transferrable across machines.
    def get_shandler(self): return List(handler for handler in self.handlers if "StreamHandler" in str(handler))
    def get_fhandler(self): return List(handler for handler in self.handlers if "FileHandler" in str(handler))
    def set_level(self, level: int, which: Literal["logger", "stream", "file", "all"] = "logger"):
        if which in {"logger", "all"}: self.setLevel(level)
        if which in {"stream", "all"}: self.get_shandler().setLevel(level)
        if which in {"file", "all"}: self.get_fhandler().setLevel(level)
    def __reduce_ex__(self, protocol: Any): _ = protocol; return self.__class__, tuple(self.specs.values())  # reduce_ex is enchanced reduce. It is lower than getstate and setstate. It uses init method to create an instance.
    def __repr__(self): return "".join([f"Logger {self.name} (level {self.level}) with handlers: \n"] + [repr(h) + f"" + "\n" for h in self.handlers])
    @staticmethod
    def get_format(sep: str = ' | ', datefmt: str = "%d %H:%M:%S"):
        _ = datefmt  # TODO: add datefmt to the format string
        return f"%(asctime)s{sep}%(name)s{sep}%(module)s{sep}%(funcName)s{sep}%(levelname)s(%(levelno)s){sep}%(message)s{sep}"  # Reference: https://docs.python.org/3/library/logging.html#logrecord-attributes logging.BASIC_FORMAT
    def manual_degug(self, path: PLike): _ = self; sys.stdout = open(path, 'w', encoding="utf-8"); sys.stdout.close(); print(f"Finished ... have a look @ \n {path}")  # all print statements will write to this file.
    @staticmethod
    def get_coloredlogs(name: Optional[str] = None, file: bool = False, file_path: OPLike = None, stream: bool = True, fmt: Optional[str] = None, sep: str = " | ", s_level: int = logging.DEBUG, f_level: int = logging.DEBUG, l_level: int = logging.DEBUG, verbose: bool = False):
        level_styles = {'spam': {'color': 'green', 'faint': True}, 'debug': {'color': 'white'}, 'verbose': {'color': 'blue'}, 'info': {'color': "green"}, 'notice': {'color': 'magenta'}, 'warning': {'color': 'yellow'}, 'success': {'color': 'green', 'bold': True},
                        'error': {'color': 'red', "faint": True, "underline": True}, 'critical': {'color': 'red', 'bold': True, "inverse": False}}  # https://coloredlogs.readthedocs.io/en/latest/api.html#available-text-styles-and-colors
        field_styles = {'asctime': {'color': 'green'}, 'hostname': {'color': 'magenta'}, 'levelname': {'color': 'black', 'bold': True}, 'path': {'color': 'blue'}, 'programname': {'color': 'cyan'}, 'username': {'color': 'yellow'}}
        if verbose: logger = install_n_import("verboselogs").VerboseLogger(name=name); logger.setLevel(l_level)  # https://github.com/xolox/python-verboselogs # verboselogs.install()  # hooks into logging module.
        else: logger = Log(name=name, dialect="logging", l_level=l_level, file=file, f_level=f_level, file_path=file_path, fmt=fmt or Log.get_format(sep), stream=stream, s_level=s_level)  # new step, not tested:
        install_n_import("coloredlogs").install(logger=logger, name="lol_different_name", level=logging.NOTSET, level_styles=level_styles, field_styles=field_styles, fmt=fmt or Log.get_format(sep), isatty=True, milliseconds=True); return logger
    def add_streamhandler(self, s_level: int = logging.DEBUG, fmt: Optional[Any] = None, module: Any = logging, name: str = "myStreamHandler"):
        shandler = module.StreamHandler(); shandler.setLevel(level=s_level); shandler.setFormatter(fmt=fmt); shandler.set_name(name); self.addHandler(shandler); print(f"    Level {s_level} stream handler for Logger `{self.name}` is created.")
    def add_filehandler(self, file_path: OPLike = None, fmt: Optional[Any] = None, f_level: int = logging.DEBUG, mode: str = "a", name: str = "myFileHandler"):
        fhandler = logging.FileHandler(filename := (P.tmpfile(name="logger_" + self.name, suffix=".log", folder="tmp_loggers") if file_path is None else P(file_path).expanduser()), mode=mode)
        fhandler.setFormatter(fmt=fmt); fhandler.setLevel(level=f_level); fhandler.set_name(name); self.addHandler(fhandler); self.file_path = filename.collapseuser(); print(f"    Level {f_level} file handler for Logger `{self.name}` is created @ " + P(filename).clickable())
    def test(self):
        List([self.debug, self.info, self.warning, self.error, self.critical]).apply(lambda func: func(f"this is a {func.__name__} message"))
        for level in range(0, 60, 5): self.log(msg=f"This is a message of level {level}", level=level)
    def get_history(self, lines: int = 200, to_html: bool = False):
        assert isinstance(self.file_path, P), f"Logger `{self.name}` does not have a file handler. Thus, no history is available."
        logs = "\n".join(self.file_path.expanduser().absolute().read_text().split("\n")[-lines:]); return install_n_import("ansi2html").Ansi2HTMLConverter().convert(logs) if to_html else logs


@dataclass
class STD:
    stdin: str
    stdout: str
    stderr: str
    returncode: int


class Response:
    @staticmethod
    def from_completed_process(cp: subprocess.CompletedProcess[str]):
        resp = Response(cmd=cp.args)
        resp.output.stdout = cp.stdout
        resp.output.stderr = cp.stderr
        resp.output.returncode = cp.returncode
        return resp
    def __init__(self, stdin: Optional[BinaryIO] = None, stdout: Optional[BinaryIO] = None, stderr: Optional[BinaryIO] = None, cmd: Optional[str] = None, desc: str = ""):
        self.std = dict(stdin=stdin, stdout=stdout, stderr=stderr)
        self.output = STD(stdin="", stdout="", stderr="", returncode=0)
        self.input = cmd
        self.desc = desc  # input command
    def __call__(self, *args: Any, **kwargs: Any) -> Optional[str]:
        _ = args, kwargs
        return self.op.rstrip() if type(self.op) is str else None
    @property
    def op(self) -> str: return self.output.stdout
    @property
    def ip(self) -> str: return self.output.stdin
    @property
    def err(self) -> str: return self.output.stderr
    @property
    def returncode(self) -> int: return self.output.returncode
    def op2path(self, strict_returncode: bool = True, strict_err: bool = False) -> Union[P, None]:
        if self.is_successful(strict_returcode=strict_returncode, strict_err=strict_err): return P(self.op.rstrip())
        return None
    def op_if_successfull_or_default(self, strict_returcode: bool = True, strict_err: bool = False) -> Optional[str]: return self.op if self.is_successful(strict_returcode=strict_returcode, strict_err=strict_err) else None
    def is_successful(self, strict_returcode: bool = True, strict_err: bool = False) -> bool:
        return ((self.returncode in {0, None}) if strict_returcode else True) and (self.err == "" if strict_err else True)
    def capture(self):
        for key in ["stdin", "stdout", "stderr"]:
            val: Optional[BinaryIO] = self.std[key]
            if val is not None and val.readable():
                self.output.__dict__[key] = val.read().decode().rstrip()
        return self
    def print_if_unsuccessful(self, desc: str = "TERMINAL CMD", capture: bool = True, strict_err: bool = False, strict_returncode: bool = False, assert_success: bool = False):
        _ = self.capture() if capture else None; success = self.is_successful(strict_err=strict_err, strict_returcode=strict_returncode)
        if assert_success: assert success, self.print(capture=False, desc=desc)
        _ = print(desc) if success else self.print(capture=False, desc=desc); return self
    def print(self, desc: str = "TERMINAL CMD", capture: bool = True):
        _ = self.capture() if capture else None; install_n_import("rich"); from rich import console; con = console.Console(); from rich.panel import Panel; from rich.text import Text  # from rich.syntax import Syntax; syntax = Syntax(my_code, "python", theme="monokai", line_numbers=True)
        tmp1 = Text("Input Command:\n"); tmp1.stylize("u bold blue"); tmp2 = Text("\nTerminal Response:\n"); tmp2.stylize("u bold blue")
        txt = tmp1 + Text(str(self.input), style="white") + tmp2 + Text("\n".join([f"{f' {idx} - {key} '}".center(40, "-") + f"\n{val}" for idx, (key, val) in enumerate(self.output.__dict__.items())]), style="white")
        con.print(Panel(txt, title=self.desc, subtitle=desc, width=150, style="bold cyan on black")); return self


class Terminal:
    def __init__(self, stdout: Optional[int] = subprocess.PIPE, stderr: Optional[int] = subprocess.PIPE, stdin: Optional[int] = subprocess.PIPE, elevated: bool = False):
        self.machine: str = __import__("platform").system()
        self.elevated: bool = elevated
        self.stdout = stdout
        self.stderr = stderr
        self.stdin = stdin
    # def set_std_system(self): self.stdout = sys.stdout; self.stderr = sys.stderr; self.stdin = sys.stdin
    def set_std_pipe(self): self.stdout = subprocess.PIPE; self.stderr = subprocess.PIPE; self.stdin = subprocess.PIPE
    def set_std_null(self): self.stdout, self.stderr, self.stdin = subprocess.DEVNULL, subprocess.DEVNULL, subprocess.DEVNULL  # Equivalent to `echo 'foo' &> /dev/null`
    def run(self, *cmds: str, shell: Optional[SHELLS] = None, check: bool = False, ip: Optional[str] = None):  # Runs SYSTEM commands like subprocess.run
        """Blocking operation. Thus, if you start a shell via this method, it will run in the main and won't stop until you exit manually IF stdin is set to sys.stdin, otherwise it will run and close quickly. Other combinations of stdin, stdout can lead to funny behaviour like no output but accept input or opposite.
        * This method is short for: res = subprocess.run("powershell command", capture_output=True, shell=True, text=True) and unlike os.system(cmd), subprocess.run(cmd) gives much more control over the output and input.
        * `shell=True` loads up the profile of the shell called so more specific commands can be run. Importantly, on Windows, the `start` command becomes availalbe and new windows can be launched.
        * `capture_output` prevents the stdout to redirect to the stdout of the script automatically, instead it will be stored in the Response object returned. # `capture_output=True` same as `stdout=subprocess.PIPE, stderr=subprocess.PIPE`"""
        my_list = list(cmds)  # `subprocess.Popen` (process open) is the most general command. Used here to create asynchronous job. `subprocess.run` is a thin wrapper around Popen that makes it wait until it finishes the task. `suprocess.call` is an archaic command for pre-Python-3.5.
        if self.machine == "Windows" and shell in {"powershell", "pwsh"}: my_list = [shell, "-Command"] + my_list  # alternatively, one can run "cmd"
        if self.elevated is False or self.is_user_admin(): resp: subprocess.CompletedProcess[str] = subprocess.run(my_list, stderr=self.stderr, stdin=self.stdin, stdout=self.stdout, text=True, shell=True, check=check, input=ip)
        else: resp = __import__("ctypes").windll.shell32.ShellExecuteW(None, "runas", sys.executable, " ".join(sys.argv), None, 1)
        return Response.from_completed_process(resp)
    def run_script(self, script: str, shell: SHELLS = "default", verbose: bool = False):
        if self.machine == "Linux": script = "#!/bin/bash" + "\n" + script  # `source` is only available in bash.
        tmp_file = P.tmpfile(name="tmp_shell_script", suffix=".ps1" if self.machine == "Windows" else ".sh", folder="tmp_scripts").write_text(script, newline={"Windows": None, "Linux": "\n"}[self.machine])
        if shell == "default":
            if self.machine == "Windows": start_cmd = "powershell"  # default shell on Windows is cmd which is not very useful. (./source is not available)
            else: start_cmd  = "."
        else: start_cmd = shell
        if verbose:
            from machineconfig.utils.utils import print_programming_script
            print_programming_script(script, lexer="shell", desc="Script to be executed:")
            import rich.progress as pb
            with pb.Progress(transient=True) as progress:
                _task = progress.add_task("Running Script", total=None)
                resp = subprocess.run([start_cmd, str(tmp_file)], stderr=self.stderr, stdin=self.stdin, stdout=self.stdout, text=True, shell=True, check=False)
        else: resp = subprocess.run([start_cmd, str(tmp_file)], stderr=self.stderr, stdin=self.stdin, stdout=self.stdout, text=True, shell=True, check=False)
        return Response.from_completed_process(resp)
    @staticmethod
    def is_user_admin() -> bool:  # adopted from: https://stackoverflow.com/questions/19672352/how-to-run-script-with-elevated-privilege-on-windows"""
        if __import__('os').name == 'nt':
            try: return __import__("ctypes").windll.shell32.IsUserAnAdmin()
            except Exception: import traceback; traceback.print_exc(); print("Admin check failed, assuming not an admin."); return False
        else: return __import__('os').getuid() == 0  # Check for root on Posix
    @staticmethod
    def run_as_admin(file: PLike, params: Any, wait: bool = False):
        proce_info = install_n_import("win32com", fromlist=["shell.shell.ShellExecuteEx"]).shell.shell.ShellExecuteEx(lpVerb='runas', lpFile=file, lpParameters=params)
        if wait: time.sleep(1)
        return proce_info
    def run_async(self, *cmds: str, new_window: bool = True, shell: Optional[str] = None, terminal: Optional[str] = None):  # Runs SYSTEM commands like subprocess.Popen
        """Opens a new terminal, and let it run asynchronously. Maintaining an ongoing conversation with another process is very hard. It is adviseable to run all
        commands in one go without interaction with an ongoing channel. Use this only for the purpose of producing a different window and humanly interact with it. Reference: https://stackoverflow.com/questions/54060274/dynamic-communication-between-main-and-subprocess-in-python & https://www.youtube.com/watch?v=IynV6Y80vws and https://www.oreilly.com/library/view/windows-powershell-cookbook/9781449359195/ch01.html"""
        if terminal is None: terminal = ""  # this means that cmd is the default console. alternative is "wt"
        if shell is None: shell = "" if self.machine == "Windows" else ""  # other options are "powershell" and "cmd". # if terminal is wt, then it will pick powershell by default anyway.
        new_window_cmd = "start" if new_window is True else ""  # start is alias for Start-Process which launches a new window.  adding `start` to the begining of the command results in launching a new console that will not inherit from the console python was launched from e.g. conda
        extra, my_list = ("-Command" if shell in {"powershell", "pwsh"} and len(cmds) else ""), list(cmds)
        if self.machine == "Windows": my_list = [new_window_cmd, terminal, shell, extra] + my_list  # having a list is equivalent to: start "ipython -i file.py". Thus, arguments of ipython go to ipython, not start.
        print("Meta.Terminal.run_async: Subprocess command: ", my_list := [item for item in my_list if item != ""])
        return subprocess.Popen(my_list, stdin=subprocess.PIPE, shell=True)  # stdout=self.stdout, stderr=self.stderr, stdin=self.stdin. # returns Popen object, not so useful for communcation with an opened terminal
    def run_py(self, script: str, wdir: OPLike = None, interactive: bool = True, ipython: bool = True, shell: Optional[str] = None, terminal: str = "", new_window: bool = True, header: bool = True):  # async run, since sync run is meaningless.
        script = ((Terminal.get_header(wdir=wdir) if header else "") + script) + ("\ntb.DisplayData.set_pandas_auto_width()\n" if terminal in {"wt", "powershell", "pwsh"} else "")
        py_script = P.tmpfile(name="tmp_python_script", suffix=".py", folder="tmp_scripts/terminal").write_text(f"""print(r'''{script}''')""" + "\n" + script)
        print(f"Script to be executed asyncronously: ", py_script.absolute().as_uri())
        shell_script = f"""
{f'cd {wdir}' if wdir is not None else ''}
{'ipython' if ipython else 'python'} {'-i' if interactive else ''} {py_script}
"""
        shell_script = P.tmpfile(name="tmp_shell_script", suffix=".sh" if self.machine == "Linux" else ".ps1", folder="tmp_scripts/shell").write_text(shell_script)
        if shell is None and self.machine == "Windows": shell = "pwsh"
        window = "start" if new_window and self.machine == "Windows" else ""
        os.system(f"{window} {terminal} {shell} {shell_script}")
    pickle_to_new_session = staticmethod(lambda obj, cmd="": Terminal().run_py(f"""path = tb.P(r'{Save.pickle(obj=obj, path=P.tmpfile(tstamp=False, suffix=".pkl"), verbose=False)}')\n obj = path.readit()\npath.delete(sure=True, verbose=False)\n {cmd}"""))
    @staticmethod
    def import_to_new_session(func: Union[None, Callable[[Any], Any]] = None, cmd: str = "", header: bool = True, interactive: bool = True, ipython: bool = True, run: bool = False, **kwargs: Any):
        load_kwargs_string = f"""kwargs = tb.P(r'{Save.pickle(obj=kwargs, path=P.tmpfile(tstamp=False, suffix=".pkl"), verbose=False)}').readit()\nkwargs.print()\n""" if kwargs else "\n"
        run_string = "\nobj(**loaded_kwargs)\n" if run else "\n"
        if callable(func) and func.__name__ != func.__qualname__:  # it is a method of a class, must be instantiated first.
            tmp = sys.modules['__main__'].__file__  # type: ignore  # pylint: disable=E1101
            assert isinstance(tmp, str), f"Cannot import a function from a module that is not a file. The module is: {tmp}"
            module = P(tmp).rel2cwd().stem if (module := func.__module__) == "__main__" else module
            load_func_string = f"import {module} as m\ninst=m.{func.__qualname__.split('.')[0]}()\nobj = inst.{func.__name__}"
        elif callable(func) and hasattr(func, "__code__"):  # it is a standalone function...
            module = P(func.__code__.co_filename)  # module = func.__module__  # fails if the function comes from main as it returns __main__.
            load_func_string = f"tb.sys.path.insert(0, r'{module.parent}')\nimport {module.stem} as m\nobj=m.{func.__name__}"
        else: load_func_string = f"""obj = tb.P(r'{Save.pickle(obj=func, path=P.tmpfile(tstamp=False, suffix=".pkl"), verbose=False)}').readit()"""
        return Terminal().run_py(load_func_string + load_kwargs_string + f"\n{cmd}\n" + run_string, header=header, interactive=interactive, ipython=ipython)  # Terminal().run_async("python", "-c", load_func_string + f"\n{cmd}\n{load_kwargs_string}\n")
    @staticmethod
    def replicate_session(cmd: str = ""): __import__("dill").dump_session(file := P.tmpfile(suffix=".pkl"), main=sys.modules[__name__]); Terminal().run_py(script=f"""path = tb.P(r'{file}')\nimport dill\nsess= dill.load_session(str(path))\npath.delete(sure=True, verbose=False)\n{cmd}""")
    @staticmethod
    def get_header(wdir: OPLike = None): return f"""\n# >> Code prepended\nimport crocodile.toolbox as tb""" + (f"""\ntb.sys.path.insert(0, r'{wdir}')""" if wdir is not None else '') + f"""\n# >> End of header, start of script passed\n"""


class SSH:  # inferior alternative: https://github.com/fabric/fabric
    def __init__(self, host: Optional[str] = None, username: Optional[str] = None, hostname: Optional[str] = None, tmate_sess: Optional[str] = None, sshkey: Optional[str] = None, pwd: Optional[str] = None, port: int = 22, ve: Optional[str] = "ve", compress: bool = False):  # https://stackoverflow.com/questions/51027192/execute-command-script-using-different-shell-in-ssh-paramiko
        self.tmate_sess = tmate_sess
        self.pwd = pwd
        self.ve = ve
        self.compress = compress  # Defaults: (1) use localhost if nothing provided.

        self.host: Optional[str] = None
        self.hostname: str
        self.username: str
        self.port: int = port
        self.proxycommand: Optional[str] = None
        import platform
        import paramiko
        # username, hostname = __import__("getpass").getuser(), platform.node()
        if isinstance(host, str):
            try:
                import paramiko.config as pconfig
                config = pconfig.SSHConfig.from_path(P.home().joinpath(".ssh/config").str)
                config_dict = config.lookup(host)
                self.hostname = config_dict["hostname"]
                self.username = config_dict["user"]
                self.host = host
                self.port = int(config_dict.get("port", port))
                sshkey = tmp[0] if type(tmp := config_dict.get("identityfile", sshkey)) is list else tmp
                self.proxycommand = config_dict.get("proxycommand", None)
                if sshkey is not None: sshkey = tmp[0] if type(tmp := config.lookup("*").get("identityfile", sshkey)) is list else tmp
            except (FileNotFoundError, KeyError):
                assert "@" in host or ":" in host, f"Host must be in the form of `username@hostname:port` or `username@hostname` or `hostname:port`, but it is: {host}"
                if "@" in host: self.username, self.hostname = host.split("@")
                else:
                    self.username = username or __import__("getpass").getuser()
                    self.hostname = host
                if ":" in self.hostname:
                    self.hostname, port_ = self.hostname.split(":")
                    self.port = int(port_)
        elif username is not None and hostname is not None:
            self.username, self.hostname = username, hostname
            self.proxycommand = None
        else:
            print(f"Provided values: host={host}, username={username}, hostname={hostname}")
            raise ValueError("Either host or username and hostname must be provided.")

        self.sshkey = str(P(sshkey).expanduser().absolute()) if sshkey is not None else None  # no need to pass sshkey if it was configured properly already
        self.ssh = paramiko.SSHClient(); self.ssh.load_system_host_keys(); self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        install_n_import("rich").inspect(Struct(host=self.host, hostname=self.hostname, username=self.username, password="***", port=self.port, key_filename=self.sshkey, ve=self.ve), value=False, title="SSHing To", docs=False, sort=False)
        sock = paramiko.ProxyCommand(self.proxycommand) if self.proxycommand is not None else None
        self.ssh.connect(hostname=self.hostname, username=self.username, password=self.pwd, port=self.port, key_filename=self.sshkey, compress=self.compress, sock=sock)  # type: ignore
        try: self.sftp = self.ssh.open_sftp()
        except Exception as err: self.sftp = None; print(f"WARNING: could not open SFTP connection to {hostname}. No data transfer is possible. Erorr faced: `{err}`")
        def view_bar(slf: Any, a: Any, b: Any): slf.total = int(b); slf.update(int(a - slf.n))  # update pbar with increment
        self.tqdm_wrap = type('TqdmWrap', (install_n_import("tqdm").tqdm,), {'view_bar': view_bar})
        self._local_distro: Optional[str] = None
        self._remote_distro: Optional[str] = None
        self._remote_machine: Optional[MACHINE] = None
        self.terminal_responses: list[Response] = []
        self.platform = platform
        self.remote_env_cmd = rf"""~/venvs/{self.ve}/Scripts/Activate.ps1""" if self.get_remote_machine() == "Windows" else rf"""source ~/venvs/{self.ve}/bin/activate"""
        self.local_env_cmd = rf"""~/venvs/{self.ve}/Scripts/Activate.ps1""" if self.platform.system() == "Windows" else rf"""source ~/venvs/{self.ve}/bin/activate"""  # works for both cmd and pwsh
    def __getstate__(self): return {attr: self.__getattribute__(attr) for attr in ["username", "hostname", "host", "tmate_sess", "port", "sshkey", "compress", "pwd", "ve"]}
    def __setstate__(self, state: dict[str, Any]): SSH(**state)
    def get_remote_machine(self) -> MACHINE:
        if self._remote_machine is None:
            if (self.run("$env:OS", verbose=False, desc="Testing Remote OS Type").op == "Windows_NT" or self.run("echo %OS%", verbose=False, desc="Testing Remote OS Type Again").op == "Windows_NT"): self._remote_machine = "Windows"
            else: self._remote_machine = "Linux"
        return self._remote_machine  # echo %OS% TODO: uname on linux
    def get_local_distro(self): self._local_distro = install_n_import("distro").name(pretty=True) if self._local_distro is None else self._local_distro; return self._local_distro
    def get_remote_distro(self):
        if self._remote_distro is None: self._remote_distro = self.run_py("print(tb.install_n_import('distro').name(pretty=True))", verbose=False).op_if_successfull_or_default() or ""
        return self._remote_distro
    def restart_computer(self): self.run("Restart-Computer -Force" if self.get_remote_machine() == "Windows" else "sudo reboot")
    def send_ssh_key(self): self.copy_from_here("~/.ssh/id_rsa.pub"); assert self.get_remote_machine() == "Windows"; self.run(P(install_n_import("machineconfig").scripts.windows.__path__.__dict__["_path"][0]).joinpath("openssh_server_add_sshkey.ps1").read_text())
    def copy_env_var(self, name: str): assert self.get_remote_machine() == "Linux"; return self.run(f"{name} = {__import__('os').environ[name]}; export {name}")
    def get_remote_repr(self, add_machine: bool = False) -> str: return f"{self.username}@{self.hostname}:{self.port}" + (f" [{self.get_remote_machine()}][{self.get_remote_distro()}]" if add_machine else "")
    def get_local_repr(self, add_machine: bool = False) -> str: return f"{__import__('getpass').getuser()}@{self.platform.node()}" + (f" [{self.platform.system()}][{self.get_local_distro()}]" if add_machine else "")
    def __repr__(self): return f"local {self.get_local_repr(add_machine=True)} >>> SSH TO >>> remote {self.get_remote_repr(add_machine=True)}"
    def run_locally(self, command: str):
        print(f"Executing Locally @ {self.platform.node()}:\n{command}")
        res = Response(cmd=command); res.output.returncode = os.system(command)
        return res
    def get_ssh_conn_str(self, cmd: str = ""): return f"ssh " + (f" -i {self.sshkey}" if self.sshkey else "") + self.get_remote_repr().replace(':', ' -p ') + (f' -t {cmd} ' if cmd != '' else ' ')
    def open_console(self, cmd: str = '', new_window: bool = True, terminal: Optional[str] = None, shell: str = "pwsh"): Terminal().run_async(*(self.get_ssh_conn_str(cmd=cmd).split(" ")), new_window=new_window, terminal=terminal, shell=shell)
    def run(self, cmd: str, verbose: bool = True, desc: str = "", strict_err: bool = False, strict_returncode: bool = False, env_prefix: bool = False) -> Response:  # most central method.
        cmd = (self.remote_env_cmd + "; " + cmd) if env_prefix else cmd
        raw = self.ssh.exec_command(cmd)
        res = Response(stdin=raw[0], stdout=raw[1], stderr=raw[2], cmd=cmd, desc=desc)  # type: ignore
        _ = res.print_if_unsuccessful(capture=True, desc=desc, strict_err=strict_err, strict_returncode=strict_returncode, assert_success=False) if not verbose else res.print(); self.terminal_responses.append(res); return res
    def run_py(self, cmd: str, desc: str = "", return_obj: bool = False, verbose: bool = True, strict_err: bool = False, strict_returncode: bool = False) -> Union[Any, Response]:
        assert '"' not in cmd, f'Avoid using `"` in your command. I dont know how to handle this when passing is as command to python in pwsh command.'
        if not return_obj: return self.run(cmd=f"""{self.remote_env_cmd}; python -c "{Terminal.get_header(wdir=None)}{cmd}\n""" + '"', desc=desc or f"run_py on {self.get_remote_repr()}", verbose=verbose, strict_err=strict_err, strict_returncode=strict_returncode)
        assert "obj=" in cmd, f"The command sent to run_py must have `obj=` statement if return_obj is set to True"
        source_file = self.run_py(f"""{cmd}\npath = tb.Save.pickle(obj=obj, path=tb.P.tmpfile(suffix='.pkl'))\nprint(path)""", desc=desc, verbose=verbose, strict_err=True, strict_returncode=True).op.split('\n')[-1]
        return self.copy_to_here(source=source_file, target=P.tmpfile(suffix='.pkl')).readit()
    def copy_from_here(self, source: PLike, target: OPLike = None, z: bool = False, r: bool = False, overwrite: bool = False, init: bool = True) -> Union[P, list[P]]:
        if init: print(f"{'>'*15} SFTP SENDING FROM `{source}` TO `{target}`")  # TODO: using return_obj do all tests required in one go.
        source_obj = P(source).expanduser().absolute()
        if not z and source_obj.is_dir():
            if r is True:
                tmp = source_obj.search("*", folders=False, r=True)
                tmp.apply(lambda file: self.copy_from_here(source=file, target=target))
                return list(tmp)
            print(f"tb.Meta.SSH Error: source is a directory! either set r=True for recursive sending or raise zip_first flag."); raise RuntimeError
        if z: print(f"ZIPPING ..."); source_obj = P(source_obj).expanduser().zip(content=True)  # .append(f"_{randstr()}", inplace=True)  # eventually, unzip will raise content flag, so this name doesn't matter.
        if target is None: target = P(source_obj).expanduser().absolute().collapseuser(strict=True); assert target.is_relative_to("~"), f"If target is not specified, source must be relative to home."
        remotepath = self.run_py(f"path=tb.P(r'{P(target).as_posix()}').expanduser()\n{'path.delete(sure=True)' if overwrite else ''}\nprint(path.parent.create())", desc=f"Creating Target directory `{P(target).parent.as_posix()}` @ {self.get_remote_repr()}", verbose=False).op or ''
        remotepath = P(remotepath.split("\n")[-1]).joinpath(P(target).name)
        print(f"SENDING `{repr(P(source_obj))}` ==> `{remotepath.as_posix()}`")
        with self.tqdm_wrap(ascii=True, unit='b', unit_scale=True) as pbar: self.sftp.put(localpath=P(source_obj).expanduser(), remotepath=remotepath.as_posix(), callback=pbar.view_bar)  # type: ignore # pylint: disable=E1129
        if z:
            _resp = self.run_py(f"""tb.P(r'{remotepath.as_posix()}').expanduser().unzip(content=False, inplace=True, overwrite={overwrite})""", desc=f"UNZIPPING {remotepath.as_posix()}", verbose=False, strict_err=True, strict_returncode=True)
            source_obj.delete(sure=True); print("\n")
        return source_obj
    def copy_to_here(self, source: PLike, target: OPLike = None, z: bool = False, r: bool = False, init: bool = True) -> P:
        if init: print(f"{'<'*15} SFTP RECEIVING FROM `{source}` TO `{target}`")
        if not z and self.run_py(f"print(tb.P(r'{source}').expanduser().absolute().is_dir())", desc="Check if source is a dir", verbose=False, strict_returncode=True, strict_err=True).op.split("\n")[-1] == 'True':
            if r:
                tmp11 = self.run_py(f"obj=tb.P(r'{source}').search(folders=False, r=True).collapseuser(strict=False)", desc="Searching for files in source", return_obj=True, verbose=False)
                assert isinstance(tmp11, List), f"Could not resolve source path {source} due to error"
                for file in tmp11:
                    self.copy_to_here(source=file.as_posix(), target=P(target).joinpath(P(file).relative_to(source)) if target else None, r=False)
            print(f"source is a directory! either set r=True for recursive sending or raise zip_first flag."); raise RuntimeError
        if z:
            tmp: Response = self.run_py(f"print(tb.P(r'{source}').expanduser().zip(inplace=False, verbose=False))", desc=f"Zipping source file {source}", verbose=False)
            tmp2 = tmp.op2path(strict_returncode=True, strict_err=True)
            if not isinstance(tmp2, P): raise RuntimeError(f"Could not zip {source} due to {tmp.err}")
            else: source = tmp2
        if target is None:
            tmpx = self.run_py(f"print(tb.P(r'{P(source).as_posix()}').collapseuser(strict=False))", desc=f"Finding default target via relative source path", strict_returncode=True, strict_err=True, verbose=False).op2path()
            if isinstance(tmpx, P): target = tmpx
            else: raise RuntimeError(f"Could not resolve target path {target} due to error")
            assert target.is_relative_to("~"), f"If target is not specified, source must be relative to home."
        target_obj = P(target).expanduser().absolute().create(parents_only=True); target_obj += '.zip' if z and '.zip' not in target_obj.suffix else ''
        if "~" in str(source):
            tmp3 = self.run_py(f"print(tb.P(r'{source}').expanduser())", desc=f"# Resolving source path address by expanding user", strict_returncode=True, strict_err=True, verbose=False).op2path()
            if isinstance(tmp3, P): source = tmp3
            else: raise RuntimeError(f"Could not resolve source path {source} due to")
        else: source = P(source)
        print(f"RECEVING `{source}` ==> `{target_obj}`")
        with self.tqdm_wrap(ascii=True, unit='b', unit_scale=True) as pbar:  # type: ignore # pylint: disable=E1129
            assert self.sftp is not None, f"Could not establish SFTP connection to {self.hostname}."
            self.sftp.get(remotepath=source.as_posix(), localpath=str(target_obj), callback=pbar.view_bar)
        if z: target_obj = target_obj.unzip(inplace=True, content=True); self.run_py(f"tb.P(r'{source.as_posix()}').delete(sure=True)", desc="Cleaning temp zip files @ remote.", strict_returncode=True, strict_err=True, verbose=False)
        print("\n"); return target_obj
    def receieve(self, source: PLike, target: OPLike = None, z: bool = False, r: bool = False) -> P:
        scout = self.run_py(cmd=f"obj=tb.SSH.scout(r'{source}', z={z}, r={r})", desc="Scouting source path on remote", return_obj=True, verbose=False)
        assert isinstance(scout, Scout)
        if not z and scout.is_dir and scout.files is not None:
            if r:
                tmp: List[P] = scout.files.apply(lambda file: self.receieve(source=file.as_posix(), target=P(target).joinpath(P(file).relative_to(source)) if target else None, r=False))
                return tmp.list[0]
            else: print(f"source is a directory! either set r=True for recursive sending or raise zip_first flag.")
        target = P(target).expanduser().absolute().create(parents_only=True) if target else scout.source_rel2home.expanduser().absolute().create(parents_only=True); target += '.zip' if z and '.zip' not in target.suffix else ''; source = scout.source_full
        with self.tqdm_wrap(ascii=True, unit='b', unit_scale=True) as pbar: self.sftp.get(remotepath=source.as_posix(), localpath=target.as_posix(), callback=pbar.view_bar)  # type: ignore # pylint: disable=E1129
        if z: target = target.unzip(inplace=True, content=True); self.run_py(f"tb.P(r'{source.as_posix()}').delete(sure=True)", desc="Cleaning temp zip files @ remote.", strict_returncode=True, strict_err=True)
        print("\n"); return target
    @staticmethod
    def scout(source: PLike, z: bool = False, r: bool = False) -> Scout:
        source_full = P(source).expanduser().absolute()
        source_rel2home = source_full.collapseuser()
        exists = source_full.exists()
        is_dir = source_full.is_dir() if exists else False
        if z and exists:
            try: source_full = source_full.zip()
            except Exception as ex:
                raise Exception(f"Could not zip {source_full} due to {ex}") from ex  # type: ignore # pylint: disable=W0719
            source_rel2home = source_full.zip()
        files = source_full.search(folders=False, r=True).apply(lambda x: x.collapseuser()) if r and exists and is_dir else None
        return Scout(source_full=source_full, source_rel2home=source_rel2home, exists=exists, is_dir=is_dir, files=files)
    def print_summary(self):   # ip=rsp.ip, op=rsp.op
        install_n_import("tabulate"); df = __import__("pandas").DataFrame.from_records(List(self.terminal_responses).apply(lambda rsp: dict(desc=rsp.desc, err=rsp.err, returncode=rsp.returncode))); print("\nSummary of operations performed:"); print(df.to_markdown())
        _ = print("\nAll operations completed successfully.\n") if ((df['returncode'].to_list()[2:] == [None] * (len(df) - 2)) and (df['err'].to_list()[2:] == [''] * (len(df) - 2))) else print("\nSome operations failed. \n"); return df


class Scheduler:
    def __init__(self, routine: Callable[['Scheduler'], Any], wait: str = "2m", max_cycles: int = 10000000000,
                 exception_handler: Optional[Callable[[Union[Exception, KeyboardInterrupt], str, 'Scheduler'], Any]] = None,
                 logger: Optional[Log] = None, sess_stats: Optional[Callable[['Scheduler'], dict[str, Any]]] = None):
        self.routine = routine  # main routine to be repeated every `wait` time period
        self.wait = str2timedelta(wait).total_seconds()  # wait period between routine cycles.
        self.logger = logger if logger is not None else Log(name="SchedLogger_" + randstr(noun=True))
        self.exception_handler = exception_handler if exception_handler is not None else self.default_exception_handler
        self.sess_start_time = datetime.now()  # to be reset at .run
        self.records = List([])
        self.cycle: int = 0
        self.max_cycles: int = max_cycles
        self.sess_stats = sess_stats or (lambda _sched: {})
    def run(self, max_cycles: Optional[int] = None, until: str = "2050-01-01"):
        self.max_cycles, self.cycle, self.sess_start_time = max_cycles or self.max_cycles, 0, datetime.now()
        while datetime.now() < datetime.fromisoformat(until) and self.cycle < self.max_cycles:  # 1- Time before Ops, and Opening Message
            time1 = datetime.now(); self.logger.warning(f"Starting Cycle {str(self.cycle).zfill(5)}. Total Run Time = {str(datetime.now() - self.sess_start_time)}. UTC {datetime.utcnow().strftime('%d %H:%M:%S')}")
            try: self.routine(self)
            except Exception as ex: self.exception_handler(ex, "routine", self)  # 2- Perform logic
            time_left = int(self.wait - (datetime.now() - time1).total_seconds())  # 4- Conclude Message
            self.cycle += 1; self.logger.warning(f"Finishing Cycle {str(self.cycle - 1).zfill(5)}. Sleeping for {self.wait}s ({time_left}s left)\n" + "-" * 100)
            try: time.sleep(time_left if time_left > 0 else 0.1)  # # 5- Sleep. consider replacing by Asyncio.sleep
            except KeyboardInterrupt as ex: self.exception_handler(ex, "sleep", self); return  # that's probably the only kind of exception that can rise during sleep.
        self.record_session_end(reason=f"Reached maximum number of cycles ({self.max_cycles})" if self.cycle >= self.max_cycles else f"Reached due stop time ({until})")
    def get_records_df(self): return __import__("pandas").DataFrame.from_records(self.records, columns=["start", "finish", "duration", "cycles", "termination reason", "logfile"] + list(self.sess_stats(self).keys()))
    def record_session_end(self, reason: str = "Not passed to function."):
        end_time = datetime.now()
        duration = end_time - self.sess_start_time
        sess_stats = self.sess_stats(self)
        self.records.append([self.sess_start_time, end_time, duration, self.cycle, reason, self.logger.file_path] + list(sess_stats.values()))
        summ = {"start time": f"{str(self.sess_start_time)}", "finish time": f"{str(end_time)}.", "duration": f"{str(duration)} | wait time {self.wait}s", "cycles ran": f"{self.cycle} | Lifetime cycles = {self.get_records_df()['cycles'].sum()}", f"termination reason": reason, "logfile": self.logger.file_path}
        tmp = Struct(summ).update(sess_stats).print(as_config=True, return_str=True, quotes=False)
        assert isinstance(tmp, str)
        self.logger.critical(f"\n--> Scheduler has finished running a session. \n" + tmp + "\n" + "-" * 100); self.logger.critical(f"\n--> Logger history.\n" + str(self.get_records_df()))
        return self
    def default_exception_handler(self, ex: Union[Exception, KeyboardInterrupt], during: str, sched: 'Scheduler') -> None:  # user decides on handling and continue, terminate, save checkpoint, etc.  # Use signal library.
        print(sched)
        self.record_session_end(reason=f"during {during}, " + str(ex))
        self.logger.exception(ex)
        raise ex


# def try_this(func, return_=None, raise_=None, run=None, handle=None, verbose=False, **kwargs):
#     try: return func()
#     except BaseException as ex:  # or Exception
#         if verbose: print(ex)
#         if raise_ is not None: raise raise_
#         if handle is not None: return handle(ex, **kwargs)
#         return run() if run is not None else return_
def generate_readme(path: PLike, obj: Any = None, desc: str = '', save_source_code: bool = True, verbose: bool = True):  # Generates a readme file to contextualize any binary files by mentioning module, class, method or function used to generate the data"""
    text: str = "# Description\n" + desc + (separator := "\n" + "-" * 50 + "\n\n")
    obj_path = P(__import__('inspect').getfile(obj)) if obj is not None else None
    path = P(path)
    if obj_path is not None:
        text += f"# Source code file generated me was located here: \n`{obj_path.collapseuser().as_posix()}`\n" + separator
        try:
            repo = install_n_import("git", "gitpython").Repo(obj_path.parent, search_parent_directories=True)
            text += f"# Last Commit\n{repo.git.execute('git log -1')}{separator}# Remote Repo\n{repo.git.execute('git remote -v')}{separator}"
            try: tmppp = obj_path.relative_to(repo.working_dir).as_posix()
            except Exception: tmppp = ""  # type: ignore
            text += f"# link to files: \n{repo.remote().url.replace('.git', '')}/tree/{repo.active_branch.commit.hexsha}/{tmppp}{separator}"
        except Exception: text += f"Could not read git repository @ `{obj_path.parent}`.\n"
    text += (f"\n\n# Code to reproduce results\n\n```python\n" + __import__("inspect").getsource(obj) + "\n```" + separator) if obj is not None else ""
    readmepath = (path / f"README.md" if path.is_dir() else (path.with_name(path.trunk + "_README.md") if path.is_file() else path)).write_text(text, encoding="utf-8")
    _ = print(f"SAVED {readmepath.name} @ {readmepath.absolute().as_uri()}") if verbose else None
    if save_source_code: P((obj.__code__.co_filename if hasattr(obj, "__code__") else None) or __import__("inspect").getmodule(obj).__file__).zip(path=readmepath.with_name(P(readmepath).trunk + "_source_code.zip"), verbose=False); print("SAVED source code @ " + readmepath.with_name("source_code.zip").absolute().as_uri()); return readmepath
def show_globals(scope: dict[str, Any], **kwargs: Any): return Struct(scope).filter(lambda k, v: "__" not in k and not k.startswith("_") and k not in {"In", "Out", "get_ipython", "quit", "exit", "sys"}).print(**kwargs)
def monkey_patch(class_inst: Any, func: Callable[[Any], Any]): setattr(class_inst.__class__, func.__name__, func)
# def capture_locals(func: Callable[[Any], Any], scope: dict[str, Any], args: Optional[Any] = None, self: Optional[str] = None, update_scope: bool = True):
#     res: dict[str, Any] = {}
#     exec(extract_code(func, args=args, self=self, include_args=True, verbose=False), scope, res)  # type: ignore
#     _ = scope.update(res) if update_scope else None; return Struct(res)
# def load_from_source_code(directory: str, obj: Optional[Any] = None, delete: bool = False):
#     P(directory).search("source_code*", r=True)[0].unzip(tmpdir := P.tmp() / timestamp(name="tmp_sourcecode"))
#     sys.path.insert(0, str(tmpdir)); sourcefile = __import__(tmpdir.find("*").stem); tmpdir.delete(sure=delete, verbose=False)
#     return getattr(sourcefile, obj) if obj is not None else sourcefile
# def extract_code(func, code: Optional[str] = None, include_args: bool = True, verbose: bool = True, copy2clipboard: bool = False, **kwargs):  # TODO: how to handle decorated functions.  add support for lambda functions.  ==> use dill for powerfull inspection"""
#     import inspect; import textwrap  # assumptions: first line could be @classmethod or @staticmethod. second line could be def(...). Then function body must come in subsequent lines, otherwise ignored.
#     raw = inspect.getsourcelines(func)[0]; lines = textwrap.dedent("".join(raw[1 + (1 if raw[0].lstrip().startswith("@") else 0):])).split("\n")
#     code_string = '\n'.join([aline if not textwrap.dedent(aline).startswith("return ") else aline.replace("return ", "return_ = ") for aline in lines])  # remove return statements if there else keep line as is.
#     title, args_header, injection_header, body_header, suffix = ((f"\n# " + f"{item} {func.__name__}".center(50, "=") + "\n") for item in ["CODE EXTRACTED FROM", "ARGS KWARGS OF", "INJECTED CODE INTO", "BODY OF", "BODY END OF"])
#     code_string = title + ((args_header + extract_arguments(func, **kwargs)) if include_args else '') + ((injection_header + code) if code is not None else '') + body_header + code_string + suffix  # added later so it has more overwrite authority.
#     _ = install_n_import("clipboard").copy(code_string) if copy2clipboard else None; print(code_string) if verbose else None; return code_string  # ready to be run with exec()
# def extract_arguments(func: Callable[[Any], Any], copy2clipboard: bool = False, **kwargs: Any):
#     ak = Struct(dict((inspect := __import__("inspect")).signature(func).parameters)).values()  # ignores self for methods automatically but also ignores args and func_kwargs.
#     res = Struct.from_keys_values(ak.name, ak.default).update(kwargs).print(as_config=True, return_str=True, justify=0, quotes=True).replace("<class 'inspect._empty'>", "None").replace("= '", "= rf'")
#     ak = inspect.getfullargspec(func); res = res + (f"{ak.varargs} = (,)\n" if ak.varargs else '') + (f"{ak.varkw} = " + "{}\n" if ak.varkw else '')  # add args = () and func_kwargs = {}
#     _ = install_n_import("clipboard").copy(res) if copy2clipboard else None; return res
# def run_cell(pointer, module=sys.modules[__name__]):
#     for cell in P(module.__file__).read_text().split("#%%"):
#         if pointer in cell.split('\n')[0]: break  # bingo
#     else: raise KeyError(f"The pointer `{pointer}` was not found in the module `{module}`")
    # print(cell); install_n_import("clipboard").copy(cell); return cell
class Experimental:
    show_globals = staticmethod(show_globals)
    monkey_patch = staticmethod(monkey_patch)
    # capture_locals = staticmethod(capture_locals)
    generate_readme = staticmethod(generate_readme)
    # load_from_source_code = staticmethod(load_from_source_code)
    # extract_code = staticmethod(extract_code)
    # extract_arguments = staticmethod(extract_arguments)
    # run_cell = staticmethod(run_cell)  # Debugging and Meta programming tools"""


if __name__ == '__main__':
    pass




"""
A collection of classes extending the functionality of Python's builtins.
author: Alex Al-Saffar
email: programmer@usa.com

This is `export file` where one can dictate what will be exposed with toolbox
"""

from crocodile.meta import Experimental, Terminal, Log, Null, Scheduler, SSH
from crocodile.meta import logging, subprocess, sys, time
# from crocodile import run
from crocodile.file_management import P, Read, Compression, Cache, encrypt, decrypt, modify_text, datetime
from crocodile.core import List, Base, Struct, Display, Save
from crocodile.core import str2timedelta, timestamp, randstr, validate_name, install_n_import

_ = str2timedelta, timestamp, randstr, validate_name, install_n_import
__ = P, Read, Compression, Cache, encrypt, decrypt, modify_text, datetime
___ = Experimental, Terminal, Log, Null, Scheduler, SSH
____ = logging, subprocess, sys, time

Path = P
L = List
S = Struct
D = Display
T = Terminal
E = Experimental
tmp = P.tmp

_1 = Base, timestamp, Save, Terminal, List, Struct, Display, P, Read, Compression, Experimental
# from crocodile.core import datetime, dt, os, sys, string, random, np, copy, dill


# _ = False
# if _:
#     from crocodile import matplotlib_management as _pm
#     from crocodile.matplotlib_management import plt, enum, FigureManager
#
#     Artist, FigurePolicy, ImShow, FigureSave = _pm.Artist, _pm.FigurePolicy, _pm.ImShow, _pm.FigureSave
#     VisibilityViewer, VisibilityViewerAuto = _pm.VisibilityViewer, _pm.VisibilityViewerAuto
#
#     _ = plt, enum, FigureManager, FigurePolicy, ImShow, FigureSave, VisibilityViewer, VisibilityViewerAuto, Artist
#
#     import pandas as pd
#     _ = pd


if __name__ == '__main__':
    # get_parser()
    pass




"""
This module contains classes that can be used to preprocess dataframes for deep learning
"""


from typing import Optional, Any, Union
import pandas as pd
import numpy as np
import numpy.typing as npt
from sklearn.preprocessing import StandardScaler, RobustScaler  # type: ignore
from sklearn.impute import SimpleImputer  # type: ignore
from sklearn.preprocessing import OneHotEncoder, OrdinalEncoder  # type: ignore
import crocodile.toolbox as tb
# from crocodile.deeplearning import DataReader


class CategoricalClipper:
    """Use pre onehot or ordinal encoding to create a new category "Other" for values that are less than a certain threshold
Note: this class does not perform imputing. I.e. Null values are ignored at fit, and passed as is at transform. Same behaviour happens with new values at transform time.
This is in contrast to behaviour of sklearn's OrdinalEncoder and OneHotEncoder, which `can` perform missing values handling if instructed to do so.
"""
    def __init__(self, thresh: float = 1.0, others_name: str = 'ClipperOther'):
        self.thresh = thresh
        self.others_name = others_name
        self.columns: Optional[list[str]] = None
        self.pre_percentage_counts: dict[str, 'pd.Series[float]'] = {}
        self.post_percentage_counts: dict[str, 'pd.Series[float]'] = {}
        self.mapper: dict[str, Any] = {}

    def __getstate__(self) -> dict[str, Any]: return self.__dict__
    def __setstate__(self, state: dict[str, Any]) -> None: self.__dict__ = state

    def fit(self, df: pd.DataFrame) -> 'CategoricalClipper':
        print("\n")
        print(f"Fitting Categorical Clipper".center(100, '-'))

        self.columns = list(df.columns)
        for col in self.columns:
            series = df[col]
            series_na = series.isna()
            if series_na.sum() != 0:
                print(f"Column `{col}` has {series.isna().sum()} NaN, NA, None values. These will be dropped before clipping.")
                print("Percentage of Nulls before dropping = ", series_na.mean() * 100)
                series = series[~series_na]
            self.pre_percentage_counts[col] = series.value_counts(normalize=True) * 100
            self.post_percentage_counts[col], self.mapper[col] = self.create_others_category(self.pre_percentage_counts[col], thresh=self.thresh, others_name=self.others_name)
            name = self.pre_percentage_counts[col].name
            self.pre_percentage_counts[col].name = "Percentage"
            self.pre_percentage_counts[col].index.name = name
            self.post_percentage_counts[col].name = "Percentage"
            self.post_percentage_counts[col].index.name = name
            print(f"`{col}` categories pre-clipper:\n{self.pre_percentage_counts[col].to_markdown()}\n\n`{col}` categories post-clipper:\n{self.post_percentage_counts[col].to_markdown()}")
        return self
    def transform(self, df: pd.DataFrame, inplace: bool = True):
        if self.columns is None: raise RuntimeError("Fit the encoder first")
        if not inplace: raise NotImplementedError
        else:
            for col in self.columns:
                series = df.loc[:, col]
                df.loc[:, col] = series.map(self.mapper[col])
            return df

    def fit_transform(self, df: pd.DataFrame, inplace: bool = True):
        self.fit(df)
        return self.transform(df, inplace=inplace)

    @staticmethod
    def create_others_category(percentage_counts: 'pd.Series[float]', thresh: float = 1.0, others_name: str = 'Other'):
        orig_caregories = percentage_counts.index.tolist()
        other_cats = percentage_counts[percentage_counts < thresh].index
        mask = percentage_counts.index.isin(other_cats)
        others_percentage = percentage_counts[mask].sum()

        percentage_counts_updated = percentage_counts[~mask]
        percentage_counts_updated[others_name] = others_percentage

        # tmp = series.value_counts(normalize=True) * 100
        # print(f"Percentage of `Other` after collation = {series.value_counts(normalize=True).Other}")
        mapper = {cat: others_name for cat in other_cats}
        mapper.update({cat: cat for cat in orig_caregories if cat not in other_cats})
        return percentage_counts_updated, mapper


class NumericalClipper:
    """
    Note: simliarly to CategoricalClipper, this class does not perform imputing. I.e. Null values are ignored at fit, and passed as is at transform.
    This behaviour is inherited from numpy's ignoring behaviour of NaN when max, min, quantile, etc. are calculated.
    """
    def __init__(self, quant_min: float, quant_max: float):
        self.columns: Optional[list[str]] = None
        self.quant_min: float = quant_min
        self.quant_max: float = quant_max
        self.value_min: dict[str, float] = {}
        self.value_max: dict[str, float] = {}

    def __getstate__(self) -> dict[str, Any]: return self.__dict__
    def __setstate__(self, state: dict[str, Any]) -> None: self.__dict__ = state

    def fit(self, df: 'pd.DataFrame'):
        self.columns = list(df.columns)
        for col in self.columns:
            series = df[col]
            self.value_min[col] = series.quantile(self.quant_min)
            self.value_max[col] = series.quantile(self.quant_max)
        self.viz()
        return self

    def viz(self) -> None:
        print("\n")
        print(f"Clipping Columns".center(100, '-'))
        print(pd.DataFrame([self.value_min, self.value_max], index=["min", "max"]).T)
        print(f"\nQuantiles used for clipping: \n- min: {self.quant_min}\n- max: {self.quant_max}")
        print(f"-" * 100)
        print("\n")

    def transform(self, df: pd.DataFrame):
        if not self.columns: raise RuntimeError("Fit the encoder first")
        for col in self.columns:
            series = df.loc[:, col]
            df.loc[:, col] = series.clip(lower=self.value_min[col], upper=self.value_max[col])
        return df

    @staticmethod
    def clip_values(y: 'pd.Series[float]', quant_min: float = 0.98, quant_max: float = 0.02):
        los_clip_uppper_hrs = y.quantile(quant_max)
        los_clip_lower_hrs = y.quantile(quant_min)
        return y.clip(lower=los_clip_lower_hrs, upper=los_clip_uppper_hrs)


class DataFrameHander:
    def __init__(self, scaler: Union[RobustScaler, StandardScaler], imputer: SimpleImputer, cols_ordinal: list[str], cols_onehot: list[str], cols_numerical: list[str],
                 encoder_onehot: OneHotEncoder,
                 encoder_ordinal: OrdinalEncoder,
                 clipper_categorical: CategoricalClipper,
                 clipper_numerical: NumericalClipper
                 ) -> None:

        self.clipper_categorical: CategoricalClipper = clipper_categorical

        self.encoder_onehot: OneHotEncoder = encoder_onehot
        self.cols_onehot: list[str] = cols_onehot
        self.encoder_ordinal: OrdinalEncoder = encoder_ordinal
        self.cols_ordinal: list[str] = cols_ordinal

        self.cols_numerical: list[str] = cols_numerical
        self.clipper_numerical: NumericalClipper = clipper_numerical
        self.imputer: SimpleImputer = imputer
        self.scaler: Union[RobustScaler, StandardScaler] = scaler

    def __getstate__(self):
        atts: list[str] = ["scaler", "imputer", "cols_numerical", "cols_ordinal", "cols_onehot", "encoder_onehot", "encoder_ordinal", "clipper_categorical", "clipper_numerical"]
        res = {}
        for att in atts:
            if hasattr(self, att):
                res[att] = getattr(self, att)
        return res

    @staticmethod
    def profile_dataframe(df: pd.DataFrame, save_path: Optional[tb.P] = None, silent: bool = False, explorative: bool = True):
        # path = data.hp.save_dir.joinpath(data.subpath, f"pandas_profile_report{appendix}.html").create(parents_only=True)
        profile_report = tb.install_n_import(library="ydata_profiling", package="ydata-profiling").ProfileReport
        # from ydata_profiling import ProfileReport as profile_report
        # profile_report = pandas_profiling.()
        # from import ProfileReport  # also try pandasgui  # import statement is kept inside the function due to collission with matplotlib
        report = profile_report(df, title="Pandas Profiling Report", explorative=explorative, silent=silent)
        if save_path is not None: report.to_file(save_path)

    @staticmethod
    def gui_dataframe(df: 'pd.DataFrame'): tb.install_n_import("pandasgui").show(df)

    def encode(self, df: pd.DataFrame, precision: str) -> pd.DataFrame:
        """Converts the dataframe to numerical format. Missing values are encoded as `pd.NA`, otherwise, encoders will fail to handle them."""
        df[self.cols_ordinal] = self.encoder_ordinal.transform(df[self.cols_ordinal])
        tmp = self.encoder_onehot.transform(df[self.cols_onehot])
        df.drop(columns=self.cols_onehot, inplace=True)
        df[self.encoder_onehot.get_feature_names_out()] = tmp
        df[self.cols_numerical] = df[self.cols_numerical].to_numpy().astype(precision)
        return df

    def impute_standardize(self, df: pd.DataFrame) -> pd.DataFrame:
        df.fillna(np.nan, inplace=True)  # SKlearn Imputer only works with Numpy's np.nan, as opposed to Pandas' pd.NA
        # all_columns = df.columns
        res = self.imputer.transform(df[self.imputer.get_feature_names_out()])
        assert isinstance(res, np.ndarray), f"Imputer returned {type(res)}, but expected np.ndarray"
        df[self.imputer.get_feature_names_out()] = res
        res = self.scaler.transform(df[self.scaler.get_feature_names_out()])
        assert isinstance(res, np.ndarray), f"Scaler returned {type(res)}, but expected np.ndarray"
        df[self.scaler.get_feature_names_out()] = res
        # return pd.DataFrame(res, columns=columns)
        return df


def check_for_nan(ip: 'npt.NDArray[Any]') -> int:
    assert len(ip.shape) == 2, f"Expected 2D array, but got {len(ip.shape)}D array"
    total_nan_count: int = 0
    for col_idx in range(ip.shape[1]):
        nan_count = np.isnan(ip[:, col_idx]).sum()
        nan_indices = np.argwhere(np.isnan(ip[:, col_idx]))
        total_nan_count += nan_count
        if nan_count > 0:
            print(f"Column {col_idx}-{ip.shape[1]} has {nan_count} NaNs")
            print(f"Locations of NaNs: {nan_indices[:5]}")
    return total_nan_count


if __name__ == "__main__":
    pass




"""plots
"""

from crocodile.core import timestamp, Save, install_n_import, validate_name
from crocodile.file_management import P, OPLike, PLike
from crocodile.meta import Terminal
from crocodile.msc.odds import Cycle

import matplotlib.pyplot as plt
# import matplotlib
from matplotlib.backend_bases import Event
from matplotlib import widgets
from matplotlib.ticker import MultipleLocator
import matplotlib.colors as mcolors  # type: ignore # noqa # pylint: disable=unused-import
from matplotlib.colors import CSS4_COLORS  # type: ignore # pylint: disable=unused-import
from matplotlib.image import AxesImage
# from matplotlib.colorbar import vi
from matplotlib import animation
from matplotlib.figure import Figure
from matplotlib.axes import Axes
import pandas as pd
import numpy as np
import numpy.typing as npt

import enum
import subprocess
import platform
from typing import Any, Optional, Literal, TypeAlias, Callable, Union, Type  # , Union


_ = install_n_import, PLike
"""TODO: add implementation https://github.com/gustavovelascoh/plot_update
"""

STREAM: TypeAlias = Literal['clear', 'accumulate', 'update']
SAVE_TYPE: TypeAlias = Literal['MPEGPipeBased', 'MPEGFileBased', 'GIFFileBased', 'GIFPipeBased']
PLT_CMAPS: list[str] = plt.colormaps()  # type: ignore


class FigurePolicy(enum.Enum):
    close_create_new = 'Close the previous figure that has the same figname and create a new fresh one'
    add_new = 'Create a new figure with same path but with added suffix'
    same = 'Grab the figure of the same path'


def assert_requirements():
    try: subprocess.check_output(['where.exe' if platform.system() == 'Windows' else 'which', 'magick'])
    except (FileNotFoundError, subprocess.CalledProcessError) as err:
        # P(r"https://www.gyan.dev/ffmpeg/builds/ffmpeg-git-full.7z").download().unzip().search()[0].rename("ffmpeg").move(r"C://")  # P("C:\\ffmpeg\\bin")  # add to PATH
        print("Installinng image magick")
        if __import__("platform").system() == "Windows":
            Terminal().run("winget install ImageMagick.ImageMagick", shell="powershell")  # gives ffmpeg as well
            print("You might need to restart your machine before PATH change impact takes place.")
        else: raise NotImplementedError from err


class GenericSave:
    stream = 'clear'
    @staticmethod
    def fig_names_to_fig_objects(fig_names: list[str]) -> list[Figure]: return [plt.figure(num=afig) for afig in fig_names]
    @staticmethod
    def get_all_fig_objects() -> list[Figure]: return [plt.figure(k) for k in plt.get_figlabels()]  # plt.get_fignums()
    def __init__(self, save_dir: OPLike = None, save_name: Optional[str] = None, watch_figs: Optional[list[Figure]] = None, max_calls: int = 2000, delay: int = 100, **kwargs: Any):
        """How to control what to be saved: you can either pass the figures to be tracked at init time, pass them dynamically at add time, or, add method will capture every figure and axis"""
        if watch_figs is None: tmp = []
        else: tmp = watch_figs
        self.watch_figs = tmp
        self.save_name = timestamp(name=save_name)
        self.save_dir = P(save_dir) if save_dir is not None else P.tmpdir(prefix="tmp_fig_save")
        self.kwargs, self.counter, self.delay, self.max = kwargs, 0, delay, max_calls
    def add(self, fignames: Optional[list[str]] = None, names: Optional[list[str]] = None, **kwargs: Any):  # generic method used at runtime, never changed.
        print(f"Saver added frame number {self.counter}", end='\r')
        self.counter += 1; plt.pause(self.delay * 0.001)
        _ = print('Turning off IO') if self.counter > self.max else None; plt.ioff()
        if fignames is not None:
            fig_objs = self.fig_names_to_fig_objects(fignames)
            self.watch_figs += fig_objs  # path sent explicitly, # None exist ==> add all else # they exist already.
        if names is None: names = [timestamp(name=str(a_figure.get_label())) for a_figure in self.watch_figs]  # individual save path, useful for PNG.
        for afig, aname in zip(self.watch_figs, names): self._save(afig, aname, **kwargs)
    def _save(self, *args: Any, **kwargs: Any): pass  # called by generic method `add`, to be implemented when subclassing.

class Null(GenericSave):  # serves as a filler.
    def __init__(self, *args: Any, **kwargs: Any): super().__init__(*args, **kwargs); self.fname = self.save_dir
    def finish(self): print(f"Nothing saved by {self}"); return self.fname
class PDF(GenericSave):  # For pdf, you just need any stream [update, clear, accumalate], preferabbly the fastest."""
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        from matplotlib.backends.backend_pdf import PdfPages
        self.fname = self.save_dir.joinpath(self.save_name + ('.pdf' if '.pdf' not in str(self.save_name) else ''))
        self.pp = PdfPages(str(self.fname))
    def _save(self, a_fig: Figure, a_name: str, bbox_inches: str = 'tight', pad_inches: float = 0.3, **kwargs: Any):
        _ = a_name
        self.pp.savefig(a_fig, bbox_inches=bbox_inches, pad_inches=pad_inches, **kwargs)
    def finish(self): print(f"Saving results ..."); self.pp.close(); print(f"SAVED PDF @", P(self.fname).absolute().as_uri()); return self
class PNG(GenericSave):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.fname = self.save_dir = self.save_dir.joinpath(self.save_name)
    def _save(self, afigure: Figure, aname: str, dpi: int = 150, **kwargs: Any):
        fname = self.save_dir.joinpath(validate_name(aname)).create(parents_only=True)
        afigure.savefig(str(fname), bbox_inches='tight', pad_inches=0.3, dpi=dpi, **kwargs)
    def finish(self): print(f"SAVED PNGs @", P(self.fname).absolute().as_uri()); return self

# class GIF(GenericSave):  # NOT RECOMMENDED, used GIFFileBased instead.
#     """This class uses ArtistAnimation: works on lines and images list attached to figure axes and Doesn't work on axes, unless you add large number of them. As such, titles are not incorporated etc (limitation).
#     Requirements: same axis must persist (If you clear the axis, nothing will be saved), only new objects are drawn inside it. Additionally, the objects drawn must not be removed, or updated, instead they should pile up in axis.
#     # usually it is smoother when adding animate=True to plot or imshow commands for GIF purpose
#     Works for images only. Add more .imshow to the same axis, and that's it. imshow will conver up previous images. For lines, it will superimpose it and will look ugly.
#     The class will automatically detect new lines by their "neo" labels and add them then hide them for the next round.
#     """
#     def __init__(self, interval: int = 100, **kwargs: Any):
#         super().__init__(**kwargs); from collections import defaultdict
#         self.container: dict[str, Any] = defaultdict(lambda: [])
#         self.interval = interval
#         self.fname = None  # determined at finish time.
#     def _save(self, afigure: Figure, aname: str, cla: bool = False, **kwargs: Any):
#         _ = aname, kwargs, cla
#         fig_list, subcontainer = self.container[afigure.get_label()], []
#         for item in FigureManager.findobj(afigure.get_label(), 'neo'): item.set_label('processed'); item.set_visible(False); subcontainer += [item]
#         fig_list.append(subcontainer)  # if you want the method coupled with cla being used in main, then it add_line is required for axes.
#     def finish(self):
#         print("Saving the GIF ....")
#         for _idx, a_fig in enumerate(self.watch_figs):
#             if ims := self.container[a_fig.get_label()]:
#                 self.fname = self.save_dir.joinpath(f'{a_fig.get_label()}_{self.save_name}.gif')
#                 ani = animation.ArtistAnimation(a_fig, ims, interval=self.interval, blit=True, repeat_delay=1000)
#                 ani.save(str(self.fname), writer=animation.PillowWriter(fps=4))  # if you don't specify the writer, it goes to ffmpeg by default then try others if that is not available, resulting in behaviours that is not consistent across machines.
#                 print(f"SAVED GIF @", P(self.fname).absolute().as_uri())
#             else: print(f"Nothing to be saved by GIF writer."); return self.fname

class GIFFileBased(GenericSave):
    def __init__(self, fps: int = 4, dpi: int = 100, bitrate: int = 1800, _type: SAVE_TYPE = 'GIFFileBased', **kwargs: Any):
        super().__init__(**kwargs); assert_requirements()
        if _type == 'GIFPipeBased':
            extension = '.gif'
            writer: Any = animation.ImageMagickFileWriter  # internally calls: matplotlib._get_executable_info("magick")
        elif _type == "GIFFileBased":
            extension = '.gif'
            writer = animation.ImageMagickWriter
        elif _type == 'MPEGFileBased':
            extension = '.mp4'
            writer = animation.FFMpegFileWriter
        elif _type == 'MPEGPipeBased':
            extension = '.mp4'
            writer = animation.FFMpegWriter
        else: raise ValueError("Unknown writer.")
        import getpass
        self.writer = writer(fps=fps, metadata={'artist': getpass.getuser()}, bitrate=bitrate)
        self.fname = self.save_dir.joinpath(self.save_name + extension)
        assert self.watch_figs, "No figure was sent during instantiation of saver, therefore the writer cannot be setup. Did you mean to use an autosaver?"
        self.writer.setup(fig=self.watch_figs[0], outfile=str(self.fname), dpi=dpi)
    def _save(self, afig: str, aname: str, **kwargs: Any) -> None:
        _ = aname, afig
        self.writer.grab_frame(**kwargs)
    def finish(self):
        print('Saving results ...')
        self.writer.finish()
        print(f"SAVED GIF @", P(self.fname).absolute().as_uri())
        return self
class GIFPipeBased(GIFFileBased):
    def __init__(self): super().__init__(_type='GIFPipeBased')
class MPEGFileBased(GIFFileBased):
    def __init__(self): super().__init__(_type='MPEGFileBased')
class MPEGPipeBased(GIFFileBased):
    def __init__(self): super().__init__(_type='MPEGPipeBased')


Saver: TypeAlias = Union[Null, PDF, PNG, GIFFileBased, GIFPipeBased, MPEGFileBased, MPEGPipeBased]

# =================================== AUTO SAVERS ==========================================
class GenericAuto(GenericSave):
    """Parses the data internally, hence requires artist with animate method implemetend. Artist needs to have .fig attribute."""
    save_type = 'auto'
    def __init__(self, saver: Saver, animator: Callable[[Any], Any], data: list['npt.NDArray[np.float64]'], names_list: Optional[list[str]] = None, **kwargs: Any):
        super().__init__(**kwargs)
        self.saver = saver
        self.plotter = None
        assert_requirements()
        self.animator = animator
        self.data = data
        self.names_list = names_list
    def animate(self):
        plt.pause(0.5)  # give time for figures to show up before updating them
        for idx, datum in __import__("tqdm").tqdm(enumerate(self.data)):
            self.animator(datum)
            self.saver.add(names=[self.names_list[idx] if self.names_list is not None else str(idx)])
        self.saver.finish()


# class GIFAuto(GenericSave):
#     def __init__(self, animator: Callable[[Any], Any], data: list['npt.NDArray[np.float64]'], interval: int = 500, extension: str = 'gif', fps: int = 4, metadata: Optional[dict[str, Any]] = None, **kwargs: Any):
#         super().__init__()
#         self.data = data
#         writer = animation.PillowWriter(fps=fps) if extension == '.mp4' else animation.FFMpegWriter(fps=fps, metadata=metadata or {}, bitrate=2500)
#         self.animator = animator
#         plt.pause(self.delay * 0.001)  # give time for figures to show up before updating them
#         self.ani = animation.FuncAnimation(fig=self.plotter.fig, func=self.plotter.animate, frames=(i for i in zip(*self.data)), interval=interval, repeat_delay=1500, fargs=None, cache_frame_data=True, save_count=10000)
#         self.fname = self.save_dir.joinpath(self.save_name + f".{extension}")
#         self.ani.save(filename=str(self.fname), writer=writer)
#         print(f"SAVED GIF @ ", P(self.fname).absolute().as_uri())
#     def animate(self) -> None:
#         plt.pause(interval=0.5)  # give time for figures to show up before updating them
#         for idx, datum in __import__("tqdm").tqdm(enumerate(self.data)):
#             self.animator(datum)
#             self.saver.add(names=[self.names_list[idx] if self.names_list is not None else str(idx)])
#         self.saver.finish()


class PDFAuto(GenericAuto):
    """Parses the data internally, hence requires artist with animate method implemetend. Artist needs to have .fig attribute."""
    save_type = 'auto'
    def __init__(self, animator: Callable[[Any], Any], data: list['npt.NDArray[np.float64]'], names_list: Optional[list[str]] = None, **kwargs: Any):
        super().__init__(**kwargs)
        self.data = data
        self.names_list = names_list
        self.animator = animator
        self.saver = PDF(**kwargs)
        assert_requirements()
        self.animate()
    def animate(self):
        plt.pause(0.5)  # give time for figures to show up before updating them
        from tqdm import tqdm
        for idx, datum in tqdm(enumerate(self.data)):
            self.animator(datum)
            self.saver.add(names=[self.names_list[idx] if self.names_list is not None else str(idx)])
        self.saver.finish()


# class PNGAuto(GenericSave):
#     def __init__(self, **kwargs: Any):
#         super().__init__(**kwargs)
#         self.saver = PNG(**kwargs)
#         self.save_dir = self.saver.save_dir
#         self.animate()
#         self.fname = self.saver.fname
#     def animate(self):
#         plt.pause(0.5)  # give time for figures to show up before updating them
#         from tqdm import tqdm
#         for idx, datum in tqdm(enumerate(self.data)):
#             self.animator(datum)
#             self.saver.add(names=[self.names_list[idx] if self.names_list is not None else str(idx)])
#         self.saver.finish()

class NullAuto(GenericAuto):
    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self.saver = Null(**kwargs)
        self.fname = self.saver.fname
        self.animate()
# class GIFFileBasedAuto(GenericAuto):
#     def __init__(self, plotter_class, data: 'npt.NDArray[np.float64]', fps: int = 4, dpi: int = 150, bitrate: int = 2500, _type: str = 'GIFFileBasedAuto', **kwargs):
#         super().__init__(**kwargs)
#         if _type == 'GIFPipeBasedAuto': writer = animation.ImageMagickFileWriter; extension = '.gif'
#         elif _type == 'MPEGFileBasedAuto': writer = animation.FFMpegFileWriter; extension = '.mp4'
#         elif _type == 'MPEGPipeBasedAuto': writer = animation.FFMpegWriter; extension = '.mp4'
#         else: raise ValueError("Unknown writer.")
#         self.saver = writer(fps=fps, metadata=dict(artist='Alex Al-Saffar'), bitrate=bitrate)
#         self.fname = self.save_dir.joinpath(self.save_name + extension)
#         self.data = lambda: (i for i in zip(*data)); self.plotter = plotter_class(*[piece[0] for piece in data], **kwargs); plt.pause(0.5); from tqdm import tqdm
#         with self.saver.saving(fig=self.plotter.fig, outfile=self.fname, dpi=dpi):
#             for datum in tqdm(self.data()): self.plotter.animate(datum); self.saver.grab_frame(); plt.pause(self.delay * 0.001)
#         print(f"SAVED GIF successfully @ {self.fname}")
# class GIFPipeBasedAuto(GIFFileBasedAuto):
#     def __init__(self, *args: Any, **kwargs: Any): super().__init__(*args, _type='GIFPipeBasedAuto', **kwargs)
# class MPEGFileBasedAuto(GIFFileBasedAuto):
#     def __init__(self, *args: Any, **kwargs: Any): super().__init__(*args, _type='MPEGFileBasedAuto', **kwargs)
# class MPEGPipeBasedAuto(GIFFileBasedAuto):
#     def __init__(self, *args: Any, **kwargs: Any): super().__init__(*args, _type='MPEGPipeBasedAuto', **kwargs)

# from matplotlib.colors import Colormap
# q = Colormap()
# mcolors.


class FigureManager:  # use as base class for Artist & Viewers to give it free animation powers.
    def __init__(self, info_loc: Optional[tuple[float, float]] = None, figpolicy: FigurePolicy = FigurePolicy.same):
        self.figpolicy = figpolicy
        self.fig: Optional[Figure] = None
        self.ax: Optional[list[Axes]] = None
        self.event: Optional[Event] = None
        self.cmaps: Cycle[str] = Cycle(PLT_CMAPS)
        self.colors: Cycle[str] = Cycle(plt.rcParams['axes.prop_cycle'].by_key()['color'])
        self.mcolors: list[str] = list(CSS4_COLORS.keys())
        tmp: list[str] = [str(item) for item in CSS4_COLORS.values()]
        self.facecolor: Cycle[str] = Cycle(iterable=tmp)
        self.cmaps.set_value('viridis')
        self.idx_cycle = Cycle([])
        self.pause: bool = False  # animation
        self.help_menu = {'_-=+[{]}\\': {'help': "Adjust Vmin Vmax. Shift + key applies change to all axes \\ toggles auto-brightness ", 'func': self.adjust_brightness},
                          "/": {'help': 'Show/Hide info text', 'func': self.text_info},
                          "h": {'help': 'Show/Hide help menu', 'func': self.show_help},
                          "tyTY": {'help': 'Change color map', 'func': self.change_cmap},
                          '<>': {'help': 'Change figure face color', 'func': self.change_facecolor},
                          "v": {'help': 'Show/Hide pixel values (Toggle)', 'func': self.show_pix_val},
                          'P': {'help': 'Pause/Proceed (Toggle)', 'func': self.pause_func},
                          'r': {'help': 'Replay', 'func': self.replay},
                          '1': {'help': 'Previous Image', 'func': self.previous}, '2': {'help': 'Next Image', 'func': self.next},
                          'S': {'help': 'Save Object', 'func': self.save},
                          'c': {'help': 'Show/Hide cursor', 'func': self.show_cursor},
                          'aA': {'help': 'Show/Hide ticks and their labels', 'func': self.show_ticks},
                          'alt+a': {'help': 'Show/Hide annotations', 'func': self.toggle_annotate}}  # IMPORTANT: add the 'alt/ctrl+key' versions of key after the key in the dictionary above, not before, otherwise the naked key version will statisfy the condition `is key in this`? in the parser.
        self.auto_brightness, self.pix_vals = False, False; self.boundaries_flag, self.annot_flag = True, False
        self.info_loc = (0.8, 0.01) if info_loc is None else info_loc
        self.message, self.message_obj, self.cursor = '', None, None
    def show_help(self, event: Any):
        default_plt = {"q ": {'help': "Quit Figure."},
                       "Ll": {'help': "change x/y scale to log and back to linear (toggle)"},
                       "Gg": {'help': "Turn on and off x and y grid respectively."},
                       "s ": {'help': "Save Figure"},
                       "f ": {'help': "Toggle Full screen"},
                       "p ": {'help': "Select / Deselect Pan"}}
        if "Keyboard shortcuts" in plt.get_figlabels(): plt.close("Keyboard shortcuts"); _ = event  # toggle
        else:
            fig = plt.figure(num="Keyboard shortcuts")
            for i, key in enumerate(self.help_menu.keys()): fig.text(0.1, 1 - 0.05 * (i + 1), f"{key:30s} {self.help_menu[key]['help']}")
            print(pd.DataFrame([[val['help'], key] for key, val in self.help_menu.items()], columns=['Action', 'Key']), "\nDefault plt Keys:\n")
            print(pd.DataFrame([[val['help'], key] for key, val in default_plt.items()], columns=['Action', 'Key']))
    # =============== EVENT METHODS ====================================
    def animate(self): pass  # a method of the artist child class that is inheriting from this class to define behaviour when user press next or previous buttons.
    def connect(self):
        if self.fig is not None: self.fig.canvas.mpl_connect('key_press_event', self.process_key); return self
    def process_key(self, event: Any):
        self.event = event  # useful for debugging.
        # event.
        for key, value in self.help_menu.items():
            if event.key in key:
                value['func'](event)  # type: ignore
                self.update_info_text(self.message)
                break
        if event.key != 'q': event.canvas.figure.canvas.draw()  # for smooth quit without throwing errors  # don't update if you want to quit.
    def toggle_annotate(self, event: Any):
        self.annot_flag = not self.annot_flag
        if event.inaxes and event.inaxes.imageevent.inaxes.images[0].set_picker(True): self.message = f"Annotation flag is toggled to {self.annot_flag}"
    # def annotate(self, event: Any, axis: Optional[Axes] = None, data: Optional[list[Any]] = None):
    #     self.event = event; e = event.mouseevent; ax = e.inaxes if axis is None else axis
    #     if not ax: return None
    #     if not hasattr(ax, 'annot_obj'):
    #         ax.annot_obj = ax.annotate("", xy=(0, 0), xytext=(-30, 30), textcoords="offset points", arrowprops=dict(arrowstyle="->", color="w", connectionstyle="arc3"),
    #                                    va="bottom", ha="left", fontsize=10, bbox=dict(boxstyle="round", fc="w"), )
    #     else: ax.annot_obj.set_visible(self.annot_flag)
    #     x, y = int(np.round(e.xdata)), int(np.round(e.ydata))
    #     z = e.inaxes.images[0].get_array()[y, x] if data is None else data[y, x]
    #     ax.annot_obj.set_text(f'x:{x}\ny:{y}\nvalue:{z:.3f}')
    #     ax.annot_obj.xy = (x, y)
    #     self.fig.canvas.draw_idle()
    def save(self, event: Any): _ = event; Save.pickle(path=P.tmpfile(name="figure_manager"), obj=self)
    def replay(self, event: Any): _ = event; self.pause = False; self.idx_cycle.set_index(0); self.message = 'Replaying'; self.animate()
    def pause_func(self, event: Any): _ = event; self.pause = not self.pause; self.message = f'Pause flag is set to {self.pause}'; self.animate()
    def previous(self, event: Any): _ = event; self.idx_cycle.previous(); self.message = f'Previous {self.idx_cycle}'; self.animate()
    def next(self, event: Any): _ = event; self.idx_cycle.next(); self.message = f'Next {self.idx_cycle}'; self.animate()
    def text_info(self, event: Any): _ = event; self.message = ''
    def change_facecolor(self, event: Any):
        assert self.fig is not None, "Figure is not defined yet."
        if event.key == '>': tmp = self.facecolor.next()
        else: tmp = self.facecolor.previous()
        self.fig.set_facecolor(tmp)
        self.message = f"Figure facecolor was set to {self.mcolors[self.facecolor.get_index()]}"
    def adjust_brightness(self, event: Any):
        ax, message = event.inaxes, "None"
        if ax is None or not ax.images: return None
        if event.key == '\\':
            self.auto_brightness = not self.auto_brightness; message = f"Auto-brightness flag is toggled to {self.auto_brightness}"
            if self.auto_brightness:
                if self.ax is not None:
                    im = self.ax[0].images[0]
                    im.norm.autoscale(im.get_array())  # type: ignore  # changes to all ims take place in animate as in ImShow and Nifti methods animate.
        vmin, vmax = ax.images[0].get_clim()
        if event.key in '-_': message = 'increase vmin'; vmin += 1
        elif event.key in '[{': message = 'decrease vmin'; vmin -= 1
        elif event.key in '=+': message = 'increase vmax'; vmax += 1
        elif event.key in ']}': message = 'decrease vmax'; vmax -= 1
        self.message = message + '  ' + str(round(vmin, 1)) + '  ' + str(round(vmax, 1))
        if event.key in '_+}{' and self.fig is not None:
            for ax in self.fig.axes:
                if ax.images: ax.images[0].set_clim((vmin, vmax))  # type: ignore
        else:
            if ax.images: ax.images[0].set_clim((vmin, vmax))
    def change_cmap(self, event: Any):
        ax = event.inaxes
        if ax is not None:
            assert self.fig is not None, "Figure is not defined yet."
            cmap = self.cmaps.next() if event.key in 'tT' else self.cmaps.previous()
            if event.key in 'TY':
                for ax in self.fig.axes:
                    for im in ax.images: im.set_cmap(cmap)  # type: ignore
            else:
                for im in ax.images:
                    im.set_cmap(cmap)
            self.message = f"Color map changed to {ax.images[0].cmap.name}"  # type: ignore
    def show_pix_val(self, event: Any):
        if (ax := event.inaxes) is not None:
            self.pix_vals = not self.pix_vals; self.message = f"Pixel values flag set to {self.pix_vals}"
            if self.pix_vals: self.show_pixels_values(ax)
            else:
                while len(ax.texts) > 0: _ = [text.remove() for text in ax.texts]
    def show_cursor(self, event: Any):
        if not (ax := event.inaxes): return None  # don't do this if c was pressed outside an axis.
        if hasattr(ax, 'cursor_'):  # is this the first time?
            if ax.cursor_ is None: ax.cursor_ = widgets.Cursor(ax=ax, vertOn=True, horizOn=True, color='red', lw=1.0)
            else: ax.cursor_ = None  # toggle the cursor.
            self.message = f'Cursor flag set to {bool(ax.cursor_)}'
        else: ax.cursor_ = None; self.show_cursor(event)  # first call
    def show_ticks(self, event: Any):
        self.boundaries_flag = not self.boundaries_flag
        if event.key == 'a' and (axis := event.inaxes): self.toggle_ticks(axis); self.message = f"Boundaries flag set to {self.boundaries_flag} in {axis}"
        elif self.ax is not None:
            for ax in self.ax: self.toggle_ticks(ax)
    # ====================== class methods ===============================
    def get_fig(self, figname: str = '', suffix: Optional[str] = None, **kwargs: Any): return FigureManager.get_fig_static(self.figpolicy, figname, suffix, **kwargs)
    def update_info_text(self, message: str):
        assert self.message_obj is not None, "Message object is not defined yet."
        _ = self.message_obj.remove() if self.message_obj else None
        assert self.fig is not None, "Figure is not defined yet."
        self.message_obj = self.fig.text(*self.info_loc, message, fontsize=8)
    @staticmethod
    def maximize_fig() -> None:
        fig = plt.get_current_fig_manager()
        if isinstance(fig, Figure): fig.full_screen_toggle()  # TODO not working appropriately ImShow.test() # The command required is backend-dependent and also OS dependent. Doesn't work if figure is not shown yet.
    def clear_axes(self) -> None:
        if self.ax is not None:
            for ax in self.ax: ax.cla()
    def transperent_fig(self) -> None:
        assert self.fig is not None, "Figure is not defined yet."
        self.fig.canvas.manager.window.attributes("-transparentcolor", "white")  # type: ignore
    def close(self): plt.close(self.fig)
    # ====================== axis helpers ========================
    @staticmethod
    def grid(ax: Axes, factor: int = 5, x_or_y: str = 'both', color: str = 'gray', alpha1: float = 0.5, alpha2: float = 0.25):
        ax.grid(which='major', axis='x', color='gray', linewidth=0.5, alpha=alpha1); ax.grid(which='major', axis='y', color='gray', linewidth=0.5, alpha=alpha1)
        if x_or_y in {'both', 'x'}: xt = ax.get_xticks(); ax.xaxis.set_minor_locator(MultipleLocator((xt[1] - xt[0]) / factor)); ax.grid(which='minor', axis='x', color=color, linewidth=0.5, alpha=alpha2)
        if x_or_y in {'both', 'y'}: yt = ax.get_yticks(); ax.yaxis.set_minor_locator(MultipleLocator((yt[1] - yt[0]) / factor)); ax.grid(which='minor', axis='y', color=color, linewidth=0.5, alpha=alpha2)
    @staticmethod
    def set_ax_size(ax: Axes, w: float, h: float, units: Literal["inches", "pixels"] = 'inches'):
        l, r, t, b = ax.figure.subplotpars.left, ax.figure.subplotpars.right, ax.figure.subplotpars.top, ax.figure.subplotpars.bottom  # type: ignore
        _ = units
        if isinstance(ax.figure, Figure): ax.figure.set_size_inches(float(w) / (r - l), float(h) / (t - b))
    @staticmethod
    def get_ax_size(ax: Axes, units: Literal["inches", "pixels"] = "inches"):
        assert ax.figure is not None, "Figure is not defined yet."
        w, h = ax.figure.get_size_inches()  # type: ignore
        if units == "pixels":
            w, h = ax.figure.get_size_inches() * ax.figure.dpi  # type: ignore
        width, height = ax.figure.subplotpars.right - ax.figure.subplotpars.left, ax.figure.subplotpars.top - ax.figure.subplotpars.bottom  # type: ignore
        return w * width, h * height
    @staticmethod
    def toggle_ticks(an_ax: Axes, state: Optional[bool] = None): _ = [line.set_visible(not line.get_visible() if state is None else state) for line in an_ax.get_yticklines() + an_ax.get_xticklines() + an_ax.get_xticklabels() + an_ax.get_yticklabels()]
    @staticmethod
    def set_ax_to_real_life_size(ax: Axes, inch_per_unit: float = 1 / 25.4): FigureManager.set_ax_size(ax, (ax.get_xlim()[1] - ax.get_xlim()[0]) * inch_per_unit, (ax.get_ylim()[1] - ax.get_ylim()[0]) * inch_per_unit)
    @staticmethod
    def show_pixels_values(ax: Axes):
        xmin, xmax = ax.get_xlim(); ymin, ymax = ax.get_ylim()
        if ymin > ymax: ymin, ymax = ymax, ymin  # default imshow settings
        for (j, i), label in np.ndenumerate(ax.images[0].get_array()):  # type: ignore
            if (xmin < i < xmax) and (ymin < j < ymax):
                ax.text(i, j, np.round(label).__int__(), ha='center', va='center', size=8)
    # ============================ matplotlib setup ==============================
    @staticmethod
    def get_nrows_ncols(num_plots: int, nrows: Optional[int] = None, ncols: Optional[int] = None) -> tuple[int, int]:
        if not nrows and not ncols:
            nrows_res, ncols_res = int(np.floor(np.sqrt(num_plots))), int(np.ceil(np.sqrt(num_plots)))
            while nrows_res * ncols_res < num_plots: ncols_res += 1
        elif not ncols and nrows:
            ncols_res = int(np.ceil(num_plots / nrows))
            nrows_res = nrows
        elif not nrows and ncols:
            nrows_res = int(np.ceil(num_plots / ncols))
            ncols_res = ncols
        else: raise ValueError("Both nrows and ncols are defined, which is not allowed.")
        return nrows_res, ncols_res
    @staticmethod
    def findobj(figname: str, obj_name: str): return plt.figure(num=figname).findobj(lambda x: x.get_label() == obj_name)
    @staticmethod
    def get_fig_static(figpolicy: FigurePolicy, figname: str = '', suffix: Optional[str] = None, **kwargs: Any):
        exist = True if figname in plt.get_figlabels() else False
        if figpolicy is FigurePolicy.same: return plt.figure(num=figname, **kwargs)
        elif figpolicy is FigurePolicy.add_new: return plt.figure(num=(timestamp(name=figname) if suffix is None else figname + suffix) if exist else figname, **kwargs)
        elif figpolicy is FigurePolicy.close_create_new:
            if exist: plt.close(figname)
            return plt.figure(num=figname, **kwargs)
    @staticmethod
    def try_figure_size() -> None:
        fig, ax = plt.subplots()
        x = np.arange(0, 100, 0.01)
        y = np.sin(x) * 100
        ax.plot(x, y); ax.axis("square"); ax.set_xlim(0, 100); ax.set_ylim(-100, 100)
        FigureManager.set_ax_to_real_life_size(ax); fig.savefig(str(P.tmp() / "trial.png"), dpi=250)
    @staticmethod
    def write(txt: str, name: str = "text", size: int = 8, **kwargs: Any) -> Figure:
        fig = plt.figure(figsize=(11.69, 8.27), num=name)
        FigureManager.maximize_fig()
        fig.clf()
        # transform=fig.transFigure
        fig.text(0.5, 0.5, txt, size=size, ha="center", va='center', **kwargs)
        return fig
    @staticmethod
    def activate_latex(size: int = 20) -> None:
        plt.rc('xtick', labelsize=size); plt.rc('ytick', labelsize=size)
        plt.rc('axes', titlesize=size); plt.rc('legend', fontsize=size / 1.5)  # rc('text', usetex=True)
        plt.rcParams['text.usetex'] = True; plt.rc('font', **{'family': 'serif', 'serif': ['Computer Modern']})
    @staticmethod
    def set_linestyles_and_markers_and_colors(test: bool = False):
        from cycler import cycler; from matplotlib import lines
        markers = list(lines.lineMarkers.keys())[:-4]  # ignore the None
        linestyles = (list(lines.lineStyles.keys())[:-3] * 10)[:len(markers)]
        colors = (plt.rcParams['axes.prop_cycle'].by_key()['color'] * 10)[:len(markers)]
        default_cycler = (cycler(linestyle=linestyles) + cycler(marker=markers) + cycler(color=colors))
        plt.rc('axes', prop_cycle=default_cycler)
        if test: _ = [plt.plot(aq + idx * 2) for idx, aq in enumerate(np.random.randn(10, 10))]


class VisibilityViewer(FigureManager):  # This is used for browsing purpose, as opposed to saving.
    # artist = ['internal', 'external'][1]  # How is the data visualized? You need artist. The artist can either be internal, as in ImShow or passed externally (especially non image data)
    parser = ['internal', 'external'][1]  # Data parsing: internal for loop to go through all the dataset passed. # Allows manual control over parsing. external for loop. It should have add method. # Manual control only takes place after the external loop is over. #TODO parallelize this.
    stream = ['clear', 'accumulate', 'update'][1]  # Streaming (Refresh mechanism): * Clear the axis. (slowest, but easy on memory) * accumulate, using visibility to hide previous axes. (Fastest but memory intensive)  * The artist has an update method. (best)
    """ This class works on hiding axes shown on a plot, so that a new plot can be drawn. Once the entire loop is finished, you can browse through the plots with the keyboard Animation linked to `animate`"""
    def __init__(self, fig: Figure):  # Downside: slow if number of axes, lines, texts, etc. are too large. In that case, it is better to avoid this viewer and plot on the fly during animation.
        super().__init__()
        self.objs_repo: list[Any] = []  # list of lists of axes and texts.
        self.fig = fig; self.connect(); self.idx_cycle = Cycle([0])
    def add(self, objs: Any) -> None:
        self.idx_cycle.expand(val=len(self.idx_cycle.list))
        self.objs_repo.append(objs)
        for obj in (self.objs_repo[-2] if len(self.objs_repo) > 1 else []): obj.set_visible(False)
        print(f"VViewer added plot number {self.idx_cycle.get_index()}", end='\r')
    def animate(self):
        for ax in self.objs_repo[self.idx_cycle.prev_index]: ax.set_visible(False)
        for ax in self.objs_repo[self.idx_cycle.get_index()]: ax.set_visible(True)
        self.fig.canvas.draw()  # type: ignore


class LineArtist(FigureManager):  # This object knows how to draw a figure from curve-type data.
    def __init__(self, ax: Optional[Axes] = None, figname: str = 'Graph', title: str = '', label: str = 'curve', style: str = 'seaborn', figpolicy: FigurePolicy = FigurePolicy.add_new, figsize: tuple[int, int] = (14, 8)):
        super().__init__(figpolicy=figpolicy)
        self.style = style
        self.title = title
        self.line: Optional[list[Any]] = None
        self.cursor = None
        self.check_b = None
        if ax is None:  # create a figure
            with plt.style.context(style=self.style):  # type: ignore
                self.fig = self.get_fig(figname, figsize=figsize)
                tmp = self.fig.subplots()
                assert tmp is not None, "Subplots failed to create axes."
                if isinstance(tmp, list): self.ax = tmp
                elif isinstance(tmp, Axes):
                    self.ax = [tmp]
        else:
            self.ax = [ax]
            self.fig: Figure = ax.figure or None  # type: ignore
        self.visibility_ax = [0.01, 0.05, 0.2, 0.15]
        self.txt: list[str] = []
        self.label: str = label
    def plot(self, *args: Any, legends: Optional[list[str]] = None, title: Optional[str] = None, **kwargs: Any):
        assert self.ax is not None, "Axes is not defined yet."
        for ax in self.ax:
            self.line = ax.plot(*args, **kwargs)
            ax.legend(legends or [])
            if title is not None: ax.set_title(title)
            ax.grid(visible=True)
    def plot_dict(self, adict: dict[str, Any], title: str = '', xlabel: str = '', ylabel: str = ''):
        for key, val in adict.items(): self.plot(val, label=key)
        assert self.ax is not None, "Axes is not defined yet."
        for ax in self.ax:
            ax.legend()
            ax.set_title(title)
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)
        return self
    def plot_twin(self, c1: Any, c2: Any, x: Optional[Any] = None, l1: str = '', l2: str = '', ax: Optional[Axes] = None):
        if ax is None:
            if self.ax is not None:
                ax = self.ax[0]
            else: raise ValueError("No axes to plot on.")
        else: raise ValueError("Not implemented yet.")
        twin_ax = ax.twinx()
        line1 = ax.plot(x or range(len(c1)), c1, color="blue", label=l1)[0]
        line2 = twin_ax.plot(c2, color="red", label=l2)[0]  # type: ignore
        twin_ax.legend([line1, line2], [l1, l2])  # type: ignore
        ax.set_ylabel(l1); twin_ax.set_ylabel(l2)
        plt.show()
    def suptitle(self, title: str):
        assert self.fig is not None, "Figure is not defined yet."
        self.txt = [str(self.fig.text(0.5, 0.98, title, ha='center', va='center', size=9))]
    def clear(self):
        assert self.fig is not None, "Figure is not defined yet."
        self.fig.clf()  # objects that use this class will demand that the artist expose this method, in addition to .plot()
    def axes(self): return self.ax  # objects that use this class will demand that the artist expose this method, in addition to .plot()
    # def visibility(self):
    #     from matplotlib.widgets import CheckButtons
    #     self.fig.subplots_adjust(left=0.3); self.visibility_ax[-1] = 0.05 * len(self.ax.lines)
    #     rax = self.fig.add_axes(self.visibility_ax)
    #     labels, visibility = [str(line.get_label()) for line in self.ax.lines], [line.get_visible() for line in self.ax.lines]
    #     self.check_b = CheckButtons(rax, labels, visibility)
    #     def func(label): index = labels.index(label); self.ax.lines[index].set_visible(not self.ax.lines[index].get_visible()); self.fig.canvas.draw()
    #     self.check_b.on_clicked(func)
    # @staticmethod
    # def styler(plot_gen):
    #     for astyle in plt.style.available:
    #         with plt.style.context(style=astyle): plot_gen(); plt.title(astyle); plt.pause(1); plt.cla()


# class VisibilityViewerAuto(VisibilityViewer):
#     # artist = ['internal', 'external'][1]
#     parser = ['internal', 'external'][0]
#     stream = ['clear', 'accumulate', 'update'][2]
#     def __init__(self, data: Optional['npt.NDArray[np.float64]'] = None, artist: Optional[Artist] = None, stream: STREAM = 'clear', save_type: Type[Saver] = Null, save_dir: OPLike = None, save_name: Optional[str] = None, delay: int = 1,
#                  titles: Optional[list[str]] = None, legends: Optional[list[str]] = None, x_labels: Optional[list[str]] = None, pause: bool = True, **kwargs: Any):
#         """data: tensor of form  NumInputsForAnimation x ArgsPerPlot (to be unstarred) x Input (possible points x signals)
#         stream: ensure that behaviour of artist is consistent with stream. When `cccumulate`, artist should create new axes whenever plot is called."""
#         self.data = data
#         self.artist: Artist = artist or Artist()
#         self.legends = [f"Plot {i}" for i in range(len(self.data))] if legends is None else legends; self.titles = titles if titles is not None else np.arange(len(self.data))
#         super().__init__(fig=self.artist.fig); _ = kwargs
#         self.saver = save_type(watch_figs=[self.artist.fig], save_dir=save_dir, save_name=save_name, delay=delay, fps=1000 / delay)
#         self.index_max, self.pause, self.stream, self.lables = len(self.data), pause, stream, x_labels
#     test = staticmethod(lambda: VisibilityViewerAuto(data=np.random.randn(10, 1, 10, 3)))
#     def animate(self):
#         for i in range(0, self.index_max):
#             self.artist.plot(*self.data[i], title=self.titles[i], legends=self.legends)  # replot the new data point on a new axis.
#             super().add(self.artist.axes()) if self.stream == 'accumulate' else self.artist.clear()
#             self.saver.add()
#             if self.pause: break
#             else: self.idx_cycle.set_index(i)
#         return self.saver.finish()  # arrived at last image and not in manual mode


class ImShow(FigureManager):
    # artist = ['internal', 'external'][0]
    # parser = ['internal', 'external'][0]
    # stream = ['clear', 'accumulate', 'update'][2]
    def __init__(self, img_tensor: 'npt.NDArray[np.float64]', sup_titles: Optional[list[str]] = None, sub_labels: Optional[list[list[str]]] = None, save_type: Type[Saver] = Null, save_name: Optional[str] = None,
                 save_dir: OPLike = None, save_kwargs: Optional[dict[str, Any]] = None,
                 subplots_adjust: Any = None, gridspec: Any = None, tight: bool = True, info_loc: Optional[tuple[float, float]] = None, nrows: Optional[int] = None, ncols: Optional[int] = None, ax: Optional[Axes] = None,
                 figsize: Optional[tuple[int, int]] = None, figname: str = 'im_show', auto_brightness: bool = True, delay: int = 200, pause: bool = False, **kwargs: Any):
        """
        :param img_tensor: size N x M x W x H [x C]  # M used spatially, N for animation.
        :param sup_titles: Titles for frames (N)
        :param sub_labels: M x N. If shape sent is M
        """
        _ = save_dir, save_kwargs, save_type, save_name
        super().__init__(figpolicy=FigurePolicy.add_new)
        n, m = len(img_tensor), len(img_tensor[0])
        self.m, self.n = m, n
        super(ImShow, self).__init__(info_loc=info_loc)
        nrows, ncols = self.get_nrows_ncols(m, nrows, ncols)
        self.img_tensor = img_tensor
        self.sub_labels = sub_labels if sub_labels is not None else [[f"{i}-{j}" for j in range(m)] for i in range(n)]
        self.sup_titles = sup_titles if sup_titles is not None else [str(item) for item in np.arange(n)]
        self.pause, self.kwargs, self.delay, self.auto_brightness = pause, kwargs, delay, auto_brightness
        self.fname = self.event = None
        self.ims: list[AxesImage] = []  # container for images.
        self.cmaps = Cycle(PLT_CMAPS); self.cmaps.set_value('viridis')
        if ax is None:
            self.fig = self.get_fig(figname=figname, figsize=(14, 9) if figsize is None else figsize, facecolor='white')
            if figsize is None: self.maximize_fig()
            if gridspec is not None:
                assert self.fig is not None
                gs = self.fig.add_gridspec(gridspec[0])
                tmp: list[Axes] = []
                for ags in gridspec[1:]:
                    assert self.fig is not None
                    qq = self.fig.add_subplot(gs[ags[0], ags[1]])
                    tmp.append(qq)
                self.ax = tmp
            else:
                tmp: list[Axes] = list(self.fig.subplots(nrows=nrows, ncols=ncols))  # type: ignore
                self.ax = tmp
        else:
            self.ax = [ax]
            fig = ax.figure
            if isinstance(fig, Figure): self.fig = fig
            raise ValueError("Figure is not defined yet.")
        # if nrows == 1 and ncols == 1: pass
        # else: self.ax = self.ax.ravel()  # make a list out of it or # make a 1D  list out of a 2D array.
        self.connect()
        # self.fig.canvas.mpl_connect("pick_event", self.annotate)
        if tight and self.fig is not None: self.fig.tight_layout()  # type: ignore
        if subplots_adjust is not None and self.fig is not None: self.fig.subplots_adjust(**subplots_adjust)  # type: ignore
        # self.saver = save_type(watch_figs=[self.fig], save_dir=save_dir, save_name=save_name, delay=delay, fps=1000 / delay, **({} if save_kwargs is None else save_kwargs))
        if isinstance(self.ax, list):  # type: ignore
            for an_ax in self.ax: self.toggle_ticks(an_ax, state=False)
        self.idx_cycle = Cycle([item for item in range(len(self.img_tensor))])  # self.animate()
    def animate(self):
        for i in range(self.idx_cycle.get_index(), self.n):
            assert self.ax is not None, "Axes is not defined yet."
            for j, (an_image, a_label, an_ax) in enumerate(zip(self.img_tensor[i], self.sub_labels[i], self.ax)):  # with zipping, the shortest of the three, will stop the loop.
                if i == 0 and self.ims.__len__() < self.m: self.ims.append(an_ax.imshow(an_image, animated=True, **self.kwargs))
                else: self.ims[j].set_data(an_image)
                if self.auto_brightness: self.ims[j].norm.autoscale(an_image)
                an_ax.set_xlabel(f'{a_label}')
            assert isinstance(self.fig, Figure)
            self.fig.suptitle(self.sup_titles[i], fontsize=8)
            # self.saver.add(names=[self.sup_titles[i]])
            if self.pause: break
            else: self.idx_cycle.set_index(i)
        if self.idx_cycle.get_index() == self.n - 1 and not self.pause:
            # self.fname = self.saver.finish()  # arrived at last image and not in manual mode
            pass
    # @staticmethod
    # def try_cmaps(im: 'npt.NDArray[np.float64]', nrows: int = 3, ncols: int = 7, **kwargs: Any):
    #      _ = ImShow(*np.array_split([plt.get_cmap(style)(im) for style in plt.colormaps()], nrows * ncols), nrows=nrows, ncols=ncols,
    #                 sub_labels=np.array_split(plt.colormaps(), nrows * ncols), **kwargs); return [plt.get_cmap(style)(im) for style in plt.colormaps()]
    # def annotate(self, event: Any, axis: Optional[Axes] = None):
    #     _ = [super().annotate(event, axis=ax, data=ax.images[0].get_array()) for ax in self.ax]
    # @staticmethod
    # def from_img_paths(paths: list[PLike], **kwargs: Any): ImShow(List(paths).apply(plt.imread), sub_labels=List(paths).apply(lambda x: P(x).stem), **kwargs)
    # @staticmethod
    # def from_complex(data: 'npt.NDArray[np.float64]', pause: bool = True, **kwargs: Any): ImShow(data.real, data.imag, np.angle(data), abs(data), labels=['Real Part', 'Imaginary Part', 'Angle in Radians', 'Absolute Value'], pause=pause, **kwargs)
    @staticmethod
    def test() -> None: ImShow(img_tensor=np.random.rand(12, 10, 80, 120, 3))  # https://ai.googleblog.com/2019/08/turbo-improved-rainbow-colormap-for.html # https://gist.github.com/mikhailov-work/ee72ba4191942acecc03fe6da94fc73f
    @staticmethod
    def resize(path: PLike, m: int, n: int):
        res = install_n_import(library="skimage", package="scikit-image").transform.resize(plt.imread(str(path)), (m, n), anti_aliasing=True)
        plt.imsave(str(path), res)


if __name__ == '__main__':
    pass





"""Crocodile Shell
"""

import argparse
import os
import random
from rich import pretty, inspect, progress, traceback, print as pprint
from rich.text import Text
from rich.console import Console

from crocodile.core import *  # type: ignore # pylint: disable=W0401,W0614 # noqa: F403,F401
from crocodile.file_management import P
import crocodile.core as core
from crocodile.file_management import *  # type: ignore # noqa: F403,F401 # pylint: disable=W0401,W0614
from crocodile.meta import *  # type: ignore # noqa: F403,F401 # pylint: disable=W0401,W0614
from crocodile import meta
# import crocodile.environment as env
from crocodile.matplotlib_management import *  # noqa: F403,F401 # type: ignore # pylint: disable=W0401,W0614
import crocodile.toolbox as tb
import numpy as np
import pandas as pd
import platform


def print_header():
    console = Console()
    pretty.install()

    _header = f"Python {platform.python_version()} in VE `{os.getenv('VIRTUAL_ENV')}` On {platform.system()}."
    _header = Text(_header)
    _header.stylize("bold blue")
    console.rule(_header, style="bold red", align="center")

    # link to tutorial or github
    _ = Text(f"Crocodile Shell")
    _.stylize("#93e6c7 on #039006")
    print(_, __import__(name='crocodile').__version__)
    print("Made with 🐍 | Built with ❤️")


tb.D.set_numpy_display()
tb.D.set_pandas_display()
D = core.Display
L = core.List
E = meta.Experimental
S = core.Struct
_ = D, L, E, S, inspect, progress, pprint, traceback, pd, np


def print_logo(logo: str):
    from crocodile.msc.ascii_art import font_box_color, charachter_color, character_or_box_color
    if platform.system() == "Windows":
        _1x = P.home().joinpath(r"AppData/Roaming/npm/figlet").exists()
        _2x = P.home().joinpath(r"AppData/Roaming/npm/lolcatjs").exists()
        _3x = P.home().joinpath(r"AppData/Local/Microsoft/WindowsApps/boxes.exe").exists()
        if _1x and _2x and _3x:
            if random.choice([True, True, False]): font_box_color(logo)
            else: charachter_color(logo)
        else:
            print(f"Missing ascii art dependencies. Install with: iwr bit.ly/cfgasciiartwindows | iex")
            _default_art = P(__file__).parent.joinpath("art").search().sample(size=1)[0]
            print(_default_art.read_text())
    else:
        def is_executable_in_path(executable_name: str) -> bool:
            path_dirs = os.environ['PATH'].split(os.pathsep)
            for path_dir in path_dirs:
                path_to_executable = os.path.join(path_dir, executable_name)
                if os.path.isfile(path_to_executable) and os.access(path_to_executable, os.X_OK): return True
            return False
        # _x1 = P.home().joinpath(".nix-profile/bin/cowsay").exists()  # P(r"/usr/games/cowsay").exists()
        # _x2 = P.home().joinpath(".nix-profile/bin/lolcat").exists()  # P(r"/usr/games/lolcat").exists()
        # _x3 = P.home().joinpath(".nix-profile/bin/boxes").exists()  # P(r"/usr/bin/boxes").exists()
        # _x4 = P.home().joinpath(".nix-profile/bin/figlet").exists()  # P(r"/usr/bin/figlet").exists()
        _x1 = is_executable_in_path("cowsay")
        _x2 = is_executable_in_path("lolcat")
        _x3 = is_executable_in_path("boxes")
        _x4 = is_executable_in_path("figlet")

        if _x1 and _x2 and _x3 and _x4:
            _dynamic_art = random.choice([True, True, True, True, False])  # classic art (True) or boxes (False)
            if _dynamic_art: character_or_box_color(logo)
            else: print(P(__file__).parent.joinpath("art").search().sample(size=1).list[0].read_text())
        else:
            print(f"Missing ascii art dependencies. Install with: curl bit.ly/cfgasciiartlinux -L | sudo bash")
            _default_art = P(__file__).parent.joinpath("art").search().sample(size=1)[0]
            print(_default_art.read_text())
    print("\n")


def build_parser():
    parser = argparse.ArgumentParser(description="Generic Parser to launch a script in a separate window.")
    parser.add_argument("--cmd", "-c", dest="cmd", help="Python command.", default="")
    parser.add_argument("--file", "-f", dest="file", help="Python command.", default="")

    args = parser.parse_args()
    # tb.Struct(args.__dict__).print(as_config=True)
    if args.file:
        code = P(args.file).read_text()
        print(code)
        exec(code, globals())  # type: ignore # pylint: disable=W0122
    elif args.cmd:
        print(args.cmd)
        exec(args.cmd, globals())  # type: ignore # pylint: disable=W0122


if __name__ == "__main__":
    print_header()
    print_logo(logo="crocodile")
    build_parser()




"""
https://docs.sqlalchemy.org/en/14/tutorial/index.html#a-note-on-the-future

"""

import time
from typing import Optional, Any, Callable

import pandas as pd

from sqlalchemy.orm import sessionmaker, Session
from sqlalchemy import create_engine, text, inspect, Engine, Connection
from sqlalchemy.engine import Inspector
from sqlalchemy.sql.schema import MetaData
from crocodile.core import Struct, Display
from crocodile.file_management import List as L, P, OPLike

_ = create_engine, text


Display.set_pandas_display()


class DBMS:
    """Implementation Philosophy:
    * Always use sqlalchemy API and avoid sql-dielect specific language.
    * Engine is provided externally. It is the end-user's business to make this engine.
    """
    def __init__(self, engine: Engine, sch: Optional[str] = None, vws: bool = False):
        self.eng: Engine = engine
        self.con: Optional[Connection] = None
        self.ses: Optional[Session] = None
        self.insp: Optional[Inspector] = None
        self.meta: Optional[MetaData] = None
        self.path = P(self.eng.url.database) if self.eng.url.database else None  # memory db

        # self.db = db
        self.sch = sch
        self.vws: bool = vws
        self.schema: Optional[L[str]] = None
        # self.tables = None
        # self.views = None
        # self.sch_tab: Optional[Struct] = None
        # self.sch_vws: Optional[Struct] = None
        self.refresh()
        # self.ip_formatter: Optional[Any] = None
        # self.db_specs: Optional[Any] = None

    def refresh(self, sch: Optional[str] = None) -> 'DBMS':  # fails if multiple schemas are there and None is specified
        self.con = self.eng.connect()
        self.ses = sessionmaker()(bind=self.eng)  # ORM style
        self.meta = MetaData()
        self.meta.reflect(bind=self.eng, schema=sch or self.sch)
        insp = inspect(subject=self.eng)
        self.insp = insp
        self.schema = L(obj_list=self.insp.get_schema_names())
        self.sch_tab: dict[str, list[str]] = {k: v for k, v in zip(self.schema.list, self.schema.apply(lambda x: insp.get_table_names(schema=x)))}  # dict(zip(self.schema, self.schema.apply(lambda x: self.insp.get_table_names(schema=x))))  #
        self.sch_vws: dict[str, list[str]] = {k: v for k, v in zip(self.schema.list, self.schema.apply(lambda x: insp.get_view_names(schema=x)))}
        return self

    def __getstate__(self) -> dict[str, Any]:
        state = self.__dict__.copy()
        del state["con"]
        del state["ses"]
        del state["meta"]
        del state["insp"]
        del state["eng"]
        if self.path:
            state['path'] = self.path.collapseuser()
        return state
    def __setstate__(self, state: dict[str, Any]) -> None:
        self.__dict__.update(state)
        self.eng = self.make_sql_engine(self.path)
        self.refresh()

    @classmethod
    def from_local_db(cls, path: OPLike = None, echo: bool = False, share_across_threads: bool = False, **kwargs: Any): return cls(engine=cls.make_sql_engine(path=path, echo=echo, share_across_threads=share_across_threads, **kwargs))
    def __repr__(self): return f"DataBase @ {self.eng}"
    def get_columns(self, table: str, sch: Optional[str] = None):
        assert self.meta is not None
        return self.meta.tables[self._get_table_identifier(table=table, sch=sch)].exported_columns.keys()
    def close(self, sleep: int = 2):
        if self.path:
            print(f"Terminating database `{self.path.as_uri() if 'memory' not in self.path else self.path}`")
        if self.con: self.con.close()
        if self.ses: self.ses.close()
        self.eng.pool.dispose()
        self.eng.dispose()
        time.sleep(sleep)
    def _get_table_identifier(self, table: str, sch: Optional[str]):
        if sch is None: sch = self.sch
        if sch is not None:
            return f"{sch}.'{table}'"
        else: return table

    @staticmethod
    def make_sql_engine(path: OPLike = None, echo: bool = False, dialect: str = "sqlite", driver: str = ["pysqlite", "DBAPI"][0], pool_size: int = 5, share_across_threads: bool = True, **kwargs: Any):
        """Establish lazy initialization with database"""
        if str(path) == "memory":
            print("Linking to in-memory database.")
            if share_across_threads:
                from sqlalchemy.pool import StaticPool  # see: https://docs.sqlalchemy.org/en/14/dialects/sqlite.html#using-a-memory-database-in-multiple-threads
                return create_engine(url=f"{dialect}+{driver}:///:memory:", echo=echo, future=True, poolclass=StaticPool, connect_args={"check_same_thread": False})
            else: return create_engine(url=f"{dialect}+{driver}:///:memory:", echo=echo, future=True, pool_size=pool_size, **kwargs)
        path = P.tmpfile(folder="tmp_dbs", suffix=".sqlite") if path is None else P(path).expanduser().absolute().create(parents_only=True)
        print(f"Linking to database at {path.as_uri()}")
        return create_engine(url=f"{dialect}+{driver}:///{path}", echo=echo, future=True, pool_size=10, **kwargs)  # echo flag is just a short for the more formal way of logging sql commands.

    # ==================== QUERIES =====================================
    def execute_as_you_go(self, *commands: str, res_func: Callable[[Any], Any] = lambda x: x.all(), df: bool = False):
        with self.eng.connect() as conn:
            result = None
            for command in commands:
                result = conn.execute(text(command))
            conn.commit()  # if driver is sqlite3, the connection is autocommitting. # this commit is only needed in case of DBAPI driver.
            return res_func(result) if not df else pd.DataFrame(res_func(result))

    def execute_begin_once(self, command: str, res_func: Callable[[Any], Any] = lambda x: x.all(), df: bool = False):
        with self.eng.begin() as conn:
            result = conn.execute(text(command))  # no need for commit regardless of driver
            result = res_func(result)
        return result if not df else pd.DataFrame(result)

    def execute(self, command: str, df: bool = False):
        with self.eng.begin() as conn: result = conn.execute(text(command))
        return result if not df else pd.DataFrame(result)

    # def execute_script(self, command: str, df: bool = False):
    #     with self.eng.begin() as conn: result = conn.executescript(text(command))
    #     return result if not df else pd.DataFrame(result)

    # ========================== TABLES =====================================
    def read_table(self, table: Optional[str] = None, sch: Optional[str] = None, size: int = 5):
        sch = sch or self.sch or 'main'
        if table is None:
            table = self.sch_tab[sch][0]
            print(f"Reading table `{table}` from schema `{sch}`")
        if self.con:
            res = self.con.execute(text(f'''SELECT * FROM {self._get_table_identifier(table, sch)} '''))
            return pd.DataFrame(res.fetchmany(size))

    def insert_dicts(self, table: str, *mydicts: dict[str, Any]) -> None:
        cmd = f"""INSERT INTO {table} VALUES """
        for mydict in mydicts: cmd += f"""({tuple(mydict)}), """
        self.execute_begin_once(cmd)

    def describe_table(self, table: str, sch: Optional[str] = None, dtype: bool = True) -> None:
        print(table.center(100, "="))
        self.refresh()
        assert self.meta is not None
        tbl = self.meta.tables[table]
        assert self.ses is not None
        count = self.ses.query(tbl).count()
        res = Struct(name=table, count=count, size_mb=count * len(tbl.exported_columns) * 10 / 1e6)
        res.print(dtype=False, as_config=True, title="TABLE DETAILS")
        dat = self.read_table(table=table, sch=sch, size=2)
        cols = self.get_columns(table, sch=sch)
        df = pd.DataFrame.from_records(dat, columns=cols)
        print("SAMPLE:\n", df)
        assert self.insp is not None
        if dtype: print("\nDETAILED COLUMNS:\n", pd.DataFrame(self.insp.get_columns(table)))
        print("\n" * 3)


if __name__ == '__main__':
    pass




"""
dl template
"""

import numpy as np
# import numpy.typing as npt
import matplotlib.pyplot as plt
import crocodile.toolbox as tb
import crocodile.deeplearning as dl
from crocodile.deeplearning import EvaluationData
from crocodile.matplotlib_management import FigureManager, Axes
import tensorflow as tf
from dataclasses import field
from typing import Optional, Any


@dl.dataclass
class HParams(dl.HParams):
    subpath: str = 'metadata/hyperparameters'  # location within model directory where this will be saved.
    name: str = field(default_factory=lambda: "model-" + tb.randstr(noun=True))
    root: tb.P = tb.P.tmp(folder="tmp_models")
    pkg_name = 'tensorflow'
    # device_name: Device=Device.gpu0
    # ===================== Data ==============================
    seed: int = 234
    shuffle: bool = True
    precision: dl.PRECISON  = 'float32'
    # ===================== Model =============================
    # depth = 3
    # ===================== Training ==========================
    test_split: float = 0.2  # test split
    learning_rate: float = 0.0005
    batch_size: int = 32
    epochs: int = 30


class DataReader(dl.DataReader):
    def __init__(self, hp: HParams, load_trianing_data: bool = False) -> None:
        specs = dl.Specs(ip_names=['x'],
                         op_names=["y"],
                         other_names=["names"],
                         ip_shapes=[],
                         op_shapes=[],
                         other_shapes=[])
        super().__init__(hp=hp, specs=specs)
        self.hp: HParams
        self.dataset: Optional[dict[str, Any]] = None
        if load_trianing_data: self.load_trianing_data()  # make sure that DataReader can be instantiated cheaply without loading data.

    def load_trianing_data(self, profile_df: bool = False):
        self.dataset = dict(x=np.random.randn(1000, 10).astype(self.hp.precision),
                            y=np.random.randn(1000, 1).astype(self.hp.precision),
                            names=np.arange(1000))
        # if profile_df: self.profile_dataframe(df=self.dataset.x)
        # self.get_pandas_profile_path()
        _ = profile_df
        self.split_the_data(data_dict=self.dataset, populate_shapes=True)

    def viz(self, eval_data: EvaluationData, ax: Optional[Axes] = None, title: str = ""):
        # _ = names
        if ax is None:
            fig, axis = plt.subplots(figsize=(14, 10))
        else:
            fig = ax.get_figure()
            axis = ax
        x = np.arange(len(eval_data.y_true[0]))
        axis.bar(x, eval_data.y_true[0].squeeze(), label='y_true', width=0.4)
        axis.bar(x + 0.4, eval_data.y_pred[0].squeeze(), label='y_pred', width=0.4)
        axis.legend()
        axis.set_title(title or 'Predicted vs True')
        FigureManager.grid(axis)
        plt.show(block=False)
        plt.pause(0.5)  # pause a bit so that the figure is displayed.
        return fig


class Model(dl.BaseModel):
    def __init__(self, hp: HParams, data: DataReader, plot: bool = False):
        super(Model, self).__init__(hp=hp, data=data)
        tf.keras.backend.set_floatx(self.hp.precision)
        self.model = self.get_model()
        self.compile()  # add optimizer and loss and metrics.
        self.build(sample_dataset=False)  # build the model (shape will be extracted from data supplied) if not passed.
        self.summary()  # print the model.
        if plot: self.plot_model()  # make sure this is not called every time the model is instantiated.

    def get_model(self):
        _ = self  # your crazy model goes here:
        m = tf.keras.Sequential([tf.keras.layers.Dense(5), tf.keras.layers.Dense(1)])
        return m

    def test(self):
        report = self.evaluate(viz=False)
        import pandas as pd
        assert isinstance(report.loss_df, pd.DataFrame)
        # loss_name = self.compiler.loss.__class__.__name__ if not hasattr(self.compiler.loss, '__name__') else self.compiler.loss.__name__
        # report.loss_df[loss_name] = report.loss_df[loss_name].round()
        df = pd.DataFrame(np.concatenate([report.y_true[0], report.y_pred[0]], axis=1).round(3), columns=["y_true", "y_pred"])
        res = pd.concat([df, report.loss_df], axis=1)
        return res


def main():
    # noinspection PyUnresolvedReferences
    # from crocodile.msc.dl_template import HParams, DataReader, Model
    hp = HParams()
    d = DataReader(hp)
    d.load_trianing_data()
    m = Model(hp, d)
    _fig, ax = plt.subplots(ncols=2)
    _res_before = m.evaluate(indices=np.arange(10).tolist(), viz_kwargs=dict(title='Before training', ax=ax[0]), viz=True)
    m.fit()
    _res_after = m.evaluate(indices=np.arange(10).tolist(), viz_kwargs=dict(title='After training', ax=ax[1]), viz=True)
    print(m.test())
    m.save_class()
    return m


if __name__ == '__main__':
    main()
    pass




"""DS
"""

import crocodile.toolbox as tb
from crocodile.cluster.remote_machine import RemoteMachine, FileManager


class Submission:
    """Sends repo, data, root_dir and write execution command."""

    @staticmethod
    def transfer_sh(rm: RemoteMachine) -> None:
        print("Using transfer.sh to send data to remote machine.")
        cloud_download_py_script = "import crocodile.toolbox as tb\n"
        # downloading repo, this takes place prior to pyscript (otherwise, its tool late as the library is loaded at the top of the pyscript already)
        if rm.config.copy_repo:
            tmp_file = tb.P(rm.job_params.repo_path_rh).expanduser().zip_n_encrypt()
            cloud_download_py_script += f"print('Downloading `{tmp_file.collapseuser()}`.')\n"
            cloud_download_py_script += f"tb.P(r'{tmp_file.share_on_cloud()}').download(folder=r'{tb.P(rm.job_params.repo_path_rh).parent}').decrypt_n_unzip()\n"
            tmp_file.delete(sure=True)
        # download data
        for _idx, item in enumerate(rm.data):
            cloud_download_py_script += f"tb.P(r'{tb.P(item).share_on_cloud()}').download(folder=r'{item.collapseuser().parent}')\n"
        # save cloud_download_script_py
        rm.file_manager.cloud_download_py_script_path.expanduser().write_text(cloud_download_py_script, encoding="utf-8")
        # modify and save shell_script to including running of cloud_download_py_script before job script.
        shell_file = rm.file_manager.shell_script_path.expanduser()
        shell_script = shell_file.read_text().replace("# EXTRA-PLACEHOLDER-POST", f"cd ~; python {rm.file_manager.cloud_download_py_script_path.rel2home().as_posix()}")
        download_url = rm.file_manager.job_root.zip().share_on_cloud()
        target = rm.file_manager.job_root.rel2home().parent.joinpath(download_url.name).as_posix()
        tmp = f"cd ~; curl -o '{target}' '{download_url.as_url_str()}'; unzip '{target}' -d {rm.file_manager.job_root.rel2home().parent.as_posix()}"
        shell_script = tmp + shell_script
        with open(file=shell_file, mode='w', newline={"Windows": None, "Linux": "\n"}[rm.ssh.get_remote_machine()], encoding="utf-8") as file: file.write(shell_script)

    @staticmethod
    def cloud(rm: RemoteMachine) -> None:
        cloud = rm.config.cloud_name
        assert cloud is not None, f"Cloud name is not specified in the config file. Please specify it in the config file."
        if rm.config.copy_repo: tb.P(rm.job_params.repo_path_rh).to_cloud(cloud=cloud, rel2home=True, zip=True, encrypt=True)
        for x in rm.data: x.to_cloud(cloud=cloud, rel2home=True)
        downloads = '\n'.join([f"cloud_copy {cloud}: '{a_path.collapseuser().as_posix()} -r" for a_path in rm.data])
        if not rm.config.copy_repo: downloads += f"""\n cloud_copy {cloud}: {tb.P(rm.job_params.repo_path_rh).collapseuser().as_posix()} -zer """
        downloads += f"\ncloud_copy {cloud}: {rm.file_manager.job_root} -zr"
        rm.file_manager.shell_script_path.expanduser().write_text(downloads + rm.file_manager.shell_script_path.expanduser().read_text(), encoding='utf-8')  # newline={"Windows": None, "Linux": "\n"}[rm.ssh.get_remote_machine()]
        tb.P(rm.file_manager.job_root).to_cloud(cloud=cloud, zip=True, rel2home=True)

    @staticmethod
    def sftp(rm: RemoteMachine) -> None:
        assert rm.ssh.sftp is not None, f"SFTP is not available for this machine `{rm.ssh}`. Consider using different `transfer_method` other than `sftp`."
        rm.ssh.run_py(f"tb.P(r'{FileManager.shell_script_path_log}').expanduser().create(parents_only=True).delete(sure=True).write_text(r'{rm.file_manager.shell_script_path.collapseuser().as_posix()}')", desc="Logging latest shell script path on remote.", verbose=False)
        if rm.config.copy_repo: rm.ssh.copy_from_here(rm.job_params.repo_path_rh, z=True, overwrite=True)
        tb.L(rm.data).apply(lambda a_path: rm.ssh.copy_from_here(a_path, z=True if tb.P(a_path).is_dir() else False, r=False, overwrite=True))
        rm.ssh.copy_from_here(rm.file_manager.job_root, z=True)


if __name__ == '__main__':
    pass




# from abc import ABC
# import crocodile.deeplearning as dl
# # import resources.toolbox as tb
# from collections import OrderedDict
# import torch as t
# # import matplotlib.pyplot as plt
# import numpy as np
# # import torch.nn
# import pandas as pd
# from typing import Optional, Any, Union

# Flatten = t.nn.Flatten


# class TorchDataReader(dl.DataReader):
#     def __init__(self, *args, **kwargs):
#         super(TorchDataReader, self).__init__(*args, **kwargs)
#         self.train_loader, self.batch = None, None
#         self.test_loader = None

#     def define_loader(self, *args):
#         s = self
#         tensors = tuple()
#         import torch.utils.data
#         for an_arg in args:
#             tensors += (t.tensor(an_arg, device=s.hp.device), )
#         tensors_dataset = t.utils.data.TensorDataset(*tensors)
#         loader = t.utils.data.DataLoader(tensors_dataset, batch_size=s.hp.batch_size)
#         batch = next(iter(loader))[0]
#         return loader, batch

#     def to_torch_tensor(self, x):
#         """.. note:: Data type is inferred from the input."""
#         return t.tensor(x).to(self.hp.device)

#     @staticmethod
#     def to_numpy(x):
#         if type(x) is not np.ndarray: return x.cpu().detach().numpy()
#         else: return x


# class PTBaseModel(dl.BaseModel, dl.ABC):

#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         self.odict = OrderedDict

#     @staticmethod
#     def check_childern_details(mod):
#         tot = 0
#         for name, layer in mod.named_children():
#             params = sum(p.numel() for p in layer.parameters())
#             print(f'Layer {name}. # Parameters = ', params)
#             tot += params
#         print(f"Total = {tot}")
#         print("-" * 20)

#     def summary(self, detailed=False):
#         print(' Summary '.center(50, '='))
#         if detailed: self.check_childern_details(self.model)
#         else:
#             print('Number of weights in the NN = ', sum(p.numel() for p in self.model.parameters()))
#             print(''.center(57, '='))

#     def save_weights(self, save_dir): t.save()
#     def save_model(self, save_dir): t.save()

#     def load_weights(self, save_dir, map_location=None):
#         if map_location is None:  # auto location.  # load to where ever the model was saved from in the first place
#             if t.cuda.is_available(): self.model.load_state_dict(t.load(save_dir.glob('*.pt').__next__()))
#             else: self.model.load_state_dict(t.load(save_dir.glob('*.pt').__next__(), map_location="cpu"))
#         else: self.model.load_state_dict(t.load(save_dir.glob('*.pt').__next__(), map_location=map_location))
#         self.model.eval()

#     def load_model(self, save_dir):  # Model class must be defined somewhere
#         self.model = t.load(self, save_dir.glob('*.pt').__next__())
#         self.model.eval()

#     def infer(self, xx):
#         self.model.eval()
#         with t.no_grad(): op = self.model(self.data.to_torch_tensor(xx))
#         return self.data.to_numpy(op)

#     def fit(self, epochs: Optional[int] = None, plot: bool = True, **kwargs):
#         """
#         """
#         if epochs is None: epochs = self.hp.epochs
#         train_losses = []
#         test_losses = []
#         print('Training'.center(100, '-'))
#         for an_epoch in range(epochs):
#             # monitor training loss
#             train_loss = 0.0
#             total_samples = 0
#             self.model.train()  # Double checking
#             for i, batch in enumerate(self.data.train_loader):
#                 _, loss, batch_length = self.train_step(batch)
#                 loss_value = loss.item()
#                 train_losses.append(loss_value)
#                 train_loss += loss_value * batch_length
#                 total_samples += batch_length
#                 if (i % 20) == 0:
#                     print(f'Accumulative loss = {train_loss}', end='\r')
#             # print avg training statistics
#             train_loss /= total_samples
#             # writer.add_scalar('training loss', train_loss, next(epoch_c))
#             test_loss = self.test(self.data.test_loader)
#             test_losses.append(test_loss[0])
#             print(f'Epoch: {an_epoch:3}/{epochs}, Training Loss: {train_loss:1.3f}, Test Loss = {test_loss[0]:1.3f}')
#         self.history.append({'loss': train_losses, 'val_loss': test_losses})
#         if plot: self.plot_loss()

#     def train_step(self, batch):
#         x, y = batch
#         self.compiler.optimizer.zero_grad()  # clear the gradients of all optimized variables
#         op = self.model(x)
#         loss = self.compiler.loss(op, y)
#         loss.backward()
#         self.compiler.optimizer.step()
#         return op, loss, len(x)

#     def test_step(self, batch):
#         with t.no_grad():
#             x, y = batch
#             op = self.model(x)
#             loss = self.compiler.loss(op, y)
#             return op, loss, len(x)

#     def test(self, loader):
#         if loader:
#             self.model.eval()
#             losses = []
#             for i, batch in enumerate(loader):
#                 prediction, loss, _ = self.test_step(batch)
#                 per_batch_losses = [loss.item()]
#                 for a_metric in self.compiler.metrics:
#                     loss = a_metric(prediction, y)
#                     per_batch_losses.append(loss.item())
#                 losses.append(per_batch_losses)
#             return [np.mean(tmp) for tmp in zip(*losses)]

#     def deploy(self, dummy_ip=None):
#         if not dummy_ip:
#             dummy_ip = t.randn_like(self.data.split.x_train[:1]).to(self.hp.device)
#         from torch import onnx
#         onnx.export(self.model, dummy_ip, 'onnx_model.onnx', verbose=True)


# class ImagesModel(PTBaseModel):
#     def __init__(self, *args): super(ImagesModel, self).__init__(*args)
#     # @tb.batcher(func_type='method')
#     def preprocess(self, images):
#         """Recieves Batch of 2D numpy input and returns tensors ready to be fed to Pytorch model.
#         """
#         images[images == 0] = self.hp.ip_mu  # To fix contrast issues, change the invalid region from 0 to 1.
#         images = images[:, None, ...]  # add channel axis first
#         images = (images - self.hp.ip_mu) / self.hp.ip_sig
#         images = self.data.to_torch_tensor(images)
#         return images

#     # @tb.batcher(func_type='method')
#     def postprocess(self, images, *args, **kwargs):
#         """  > cpu > squeeeze > np > undo norm
#         Recieves tensors from model and returns numpy images. """
#         images = self.data.to_numpy(images)
#         images = images[:, 0, ...]  # removing channel axis.
#         images = (images * self.hp.op_sig) + self.hp.op_mu
#         return images

#     @staticmethod
#     def make_channel_last(images): return images.transpose((0, 2, 3, 1)) if len(images.shape) == 4 else images.transpose((1, 2, 0))
#     @staticmethod
#     def make_channel_first(images): return images.transpose((0, 3, 1, 2)) if len(images.shape) == 4 else images.transpose((2, 0, 1))


# def check_shapes(module, ip):
#     """Used to check sizes after each layer in a Pytorch model. Use the function to mimic every call in the forwatd
#     method of a Pytorch model.

#     :param module: a module used in a single step of forward method of a model
#     :param ip: a random tensor of appropriate input size for module
#     :return: output tensor, and prints sizes along the pipeline
#     """
#     print(getattr(module, '_get_name')().center(50, '-'))
#     op = 'Input shape'
#     print(f'{0:2}- {op:20s}, {ip.shape}')
#     named_childern = list(module.named_children())
#     if len(named_childern) == 0:  # a single layer, rather than nn.Module subclass
#         named_childern = list(module.named_modules())
#     for idx, (a_name, a_layer) in enumerate(named_childern):
#         if idx == 0:
#             with t.no_grad(): op = a_layer(ip)
#         else:
#             with t.no_grad(): op = a_layer(op)
#         print(f'{idx + 1:2}- {a_name:20s}, {op.shape}')
#     print("Stats on output data for random normal input:")
#     print(pd.DataFrame(TorchDataReader.to_numpy(op).flatten()).describe())
#     return op


# class Accuracy(object):
#     """ Useful for Pytorch saved_models. Stolen from TF-Keras.
#         Measures the accuracy in a classifier. Accepts logits input, will be sigmoided inside.
#     """

#     def __init__(self):
#         self.counter = 0.0
#         self.total = 0.0

#     def reset(self):
#         self.counter = 0.0
#         self.total = 0.0

#     def update(self, pred, correct):
#         """Used during training process to find overall accuracy through out an epoch
#         """
#         self.counter += len(correct)
#         tmpo = t.tensor(t.round(t.sigmoid(pred.squeeze())) == correct.squeeze().round()).mean()
#         self.total += tmpo * len(correct)
#         return tmpo

#     @staticmethod
#     def measure(pred, correct):
#         """ This method measures the accuracy for once. Useful at test time_produced, rather than training time_produced.
#         """
#         return t.tensor(t.round(t.sigmoid(pred.squeeze())) == correct.squeeze().round()).mean()

#     def result(self): return self.total / self.counter


# class View(t.nn.Module, ABC):
#     def __init__(self, shape):
#         super(View, self).__init__()
#         self.shape = shape
#     def forward(self, xx): return xx.view(*self.shape)


# class MeanSquareError:
#     """Only for Pytorch models"""
#     def __init__(self, x_mask=1, y_mask=1):
#         self.name = 'MeanSquaredError'
#         self.x_mask = x_mask
#         self.y_mask = y_mask
#     def __call__(self, x, y):
#         x = self.x_mask * x
#         y = self.y_mask * y
#         return ((x - y) ** 2).mean(tuple(range(1, len(x.shape)))).mean(0)
#         # avoid using dim and axis keywords to make it work for both numpy and torch tensors.


# class MeanAbsoluteError:
#     """
#     Only for Pytorch models
#     """
#     def __init__(self, x_mask=1, y_mask=1):
#         self.name = 'L1Loss'
#         self.x_mask = x_mask
#         self.y_mask = y_mask
#     def __call__(self, x, y):
#         x = self.x_mask * x
#         y = self.y_mask * y
#         return (abs(x - y)).mean()
